"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.interface.virtual_backend import VirtualBackend

class QiskitBackend(VirtualBackend):
    """
    This class instantiates a VirtualBackend using a Qiskit backend.
    This allows easy access to Qiskit backends through the qrisp interface.


    Parameters
    ----------
    backend : Qiskit backend object, optional
        A Qiskit backend object, which runs QuantumCircuits. The default is
        ``AerSimulator()``.
    port : int, optional
        The port to listen. The default is 8079.

    Examples
    --------

    We evaluate a :ref:`QuantumFloat` multiplication on the Aer simulator.

    >>> from qrisp import QuantumFloat
    >>> from qrisp.interface import QiskitBackend
    >>> from qiskit_aer import AerSimulator
    >>> example_backend = QiskitBackend(backend = AerSimulator())
    >>> qf = QuantumFloat(4)
    >>> qf[:] = 3
    >>> res = qf*qf
    >>> res.get_measurement(backend = example_backend)
    {9: 1.0}


    """

    def __init__(self, backend=None, port=None):
        if backend is None:

            try:
                from qiskit_aer import AerSimulator

                backend = AerSimulator()
            except ImportError:
                from qiskit.providers.basic_provider import BasicProvider

                backend = BasicProvider().get_backend("basic_simulator")

        # Create the run method
        def run(qasm_str, shots=None, token=""):
            if shots is None:
                shots = 100000
            # Convert to qiskit
            from qiskit import QuantumCircuit

            qiskit_qc = QuantumCircuit.from_qasm_str(qasm_str)

            # Make circuit with one monolithic register
            new_qiskit_qc = QuantumCircuit(len(qiskit_qc.qubits), len(qiskit_qc.clbits))
            for instr in qiskit_qc:
                new_qiskit_qc.append(
                    instr.operation,
                    [qiskit_qc.qubits.index(qb) for qb in instr.qubits],
                    [qiskit_qc.clbits.index(cb) for cb in instr.clbits],
                )

            from qiskit import transpile

            qiskit_qc = transpile(new_qiskit_qc, backend=backend)
            # Run Circuit on the Qiskit backend
            qiskit_result = backend.run(qiskit_qc, shots=shots).result().get_counts()
            # Remove the spaces in the qiskit result keys
            result_dic = {}
            import re

            for key in qiskit_result.keys():
                counts_string = re.sub(r"\W", "", key)
                result_dic[counts_string] = qiskit_result[key]

            return result_dic

        # Call VirtualBackend constructor
        if isinstance(backend.name, str):
            name = backend.name
        else:
            name = backend.name()

        super().__init__(run, port=port)

def VirtualQiskitBackend(*args, **kwargs):
    import warnings
    warnings.warn("VirtualQiskitBackend will be deprecated in a future release of Qrisp. Use QiskitBackend instead.")
    return QiskitBackend(*args, **kwargs)

class QiskitRuntimeBackend(VirtualBackend):
    """
    This class instantiates a VirtualBackend using a Qiskit Runtime backend.
    This allows easy access to Qiskit Runtime backends through the qrisp interface.
    It is imporant to close the session after the execution of the algorithm
    (as reported in the example below).

    Parameters
    ----------
    backend : str, optional
        A string associated to the name of a Qiskit Runtime Backend.
        The default one is the "ibmq_qasm_simulator", but it also
        possible to provide other ones like "simulator_statevector"
        or real backends from the Qiskit Runtime.
    port : int, optional
        The port to listen. The default is 8079.
    token : str
        The token is necessary to create correctly the Qiskit Runtime
        service and be able to run algorithms on their backends.

    Examples
    --------

    >>> from qrisp import QuantumFloat
    >>> from qrisp.interface import QiskitRuntimeBackend
    >>> example_backend = QiskitRuntimeBackend(backend="simulator_statevector",token='YOUR_TOKEN')
    >>> qf = QuantumFloat(4)
    >>> qf[:] = 3
    >>> res = qf*qf
    >>> result=res.get_measurement(backend = example_backend)
    >>> print(results)
    >>> example_backend.close_session()
    {9: 1.0}

    """

    def __init__(self, backend=None, port=8079, token=""):
        from qiskit_ibm_runtime import QiskitRuntimeService, Sampler, Session, Estimator

        service = QiskitRuntimeService(channel="ibm_quantum", token=token)
        if backend is None:
            backend = service.get_backend("ibmq_qasm_simulator")
        else:
            backend = service.get_backend(backend)

        session = Session(service, backend)
        sampler = Sampler(session=session)

        # Create the run method
        def run(qasm_str, shots=None, token=""):
            if shots is None:
                shots = 100000
            # Convert to qiskit
            from qiskit import QuantumCircuit

            qiskit_qc = QuantumCircuit.from_qasm_str(qasm_str)

            from qiskit import transpile

            qiskit_qc = transpile(qiskit_qc, backend=backend)
            # Run Circuit with the Sampler Primitive
            qiskit_result = (
                sampler.run(qiskit_qc, shots=shots)
                .result()
                .quasi_dists[0]
                .binary_probabilities()
            )
            qiskit_result.update(
                (key, round(value * shots)) for key, value in qiskit_result.items()
            )

            # Remove the spaces in the qiskit result keys
            result_dic = {}
            import re

            for key in qiskit_result.keys():
                counts_string = re.sub(r"\W", "", key)
                result_dic[counts_string] = qiskit_result[key]

            return result_dic

        super().__init__(run, port=port)

    def close_session(self):
        """
        Method to call in order to close the session started by the init method.
        """
        self.session.close()



"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import math

import numpy as np
from numba import njit

from qrisp.core import QuantumVariable, QuantumArray
from qrisp.core.compilation import qompiler


def get_measurement(
    hamiltonian,
    qarg,
    precision=0.01,
    backend=None,
    shots=1000000,
    compile=True,
    compilation_kwargs={},
    subs_dic={},
    precompiled_qc=None,
    diagonalisation_method="ommuting_qw",
    measurement_data=None # measurement settings
    ):
    r"""
    This method returns the expected value of a Hamiltonian for the state of a quantum argument.

    Parameters
    ----------
    qarg : QuantumVariable, QuantumArray or list[QuantumVariable]
        The quantum argument to evaluate the Hamiltonian on.
    precision: float, optional
        The precision with which the expectation of the Hamiltonian is to be evaluated.
        The default is 0.01. The number of shots scales quadratically with the inverse precision.
    backend : BackendClient, optional
        The backend on which to evaluate the quantum circuit. The default can be
        specified in the file default_backend.py.
    shots : integer, optional
        The maximum amount of shots to evaluate the expectation of the Hamiltonian. 
        The default is 1000000.
    compile : bool, optional
        Boolean indicating if the .compile method of the underlying QuantumSession
        should be called before. The default is True.
    compilation_kwargs  : dict, optional
        Keyword arguments for the compile method. For more details check
        :meth:`QuantumSession.compile <qrisp.QuantumSession.compile>`. The default
        is ``{}``.
    subs_dic : dict, optional
        A dictionary of Sympy symbols and floats to specify parameters in the case
        of a circuit with unspecified, :ref:`abstract parameters<QuantumCircuit>`.
        The default is {}.
    precompiled_qc : QuantumCircuit, optional
            A precompiled quantum circuit.
    measurement_data : QubitOperatorMeasurement
        Cached data to accelerate the measurement procedure. Automatically generated by default.

    Raises
    ------
    Exception
        If the containing QuantumSession is in a quantum environment, it is not
        possible to execute measurements.

    Returns
    -------
    float
        The expected value of the Hamiltonian.

    """

    from qrisp import QuantumSession, merge

    if isinstance(qarg,QuantumVariable):
        if qarg.is_deleted():
            raise Exception("Tried to get measurement from deleted QuantumVariable")
        qs = qarg.qs
            
    elif isinstance(qarg,QuantumArray):
        for qv in qarg.flatten():
            if qv.is_deleted():
                raise Exception(
                    "Tried to measure QuantumArray containing deleted QuantumVariables"
                )
        qs = qarg.qs
    elif isinstance(qarg,list):
        qs = QuantumSession()
        for arg in qarg:
            if isinstance(arg, QuantumVariable) and qv.is_deleted():
                raise Exception(
                    "Tried to measure QuantumArray containing deleted QuantumVariables"
                ) 
            merge(qs,arg)

    if backend is None:
        if qs.backend is None:
            from qrisp.default_backend import def_backend

            backend = def_backend
        else:
            backend = qarg.qs.backend

    if len(qs.env_stack) != 0:
        raise Exception("Tried to get measurement within open environment")


    # Copy circuit in over to prevent modification
    if precompiled_qc is None:        
        if compile:
            qc = qompiler(
                qs, **compilation_kwargs
            )
        else:
            qc = qs.copy()
    else:
        qc = precompiled_qc.copy()

    # Bind parameters
    if subs_dic:
        qc = qc.bind_parameters(subs_dic)
        from qrisp.core.compilation import combine_single_qubit_gates
        qc = combine_single_qubit_gates(qc)

    qc = qc.transpile()
    
    if measurement_data is None:
        measurement_data = QubitOperatorMeasurement(hamiltonian, diagonalisation_method = diagonalisation_method)
    
    return measurement_data.get_measurement(qc, qarg, precision, backend)
    

class QubitOperatorMeasurement:
    
    def __init__(self, qubit_op, diagonalisation_method="commuting_qw"):
        
        n = qubit_op.find_minimal_qubit_amount()
        hamiltonian = qubit_op.hermitize()
        hamiltonian = hamiltonian.eliminate_ladder_conjugates()
        
        if diagonalisation_method=="commuting_qw":
            self.groups = hamiltonian.commuting_qw_groups()
        else:
            self.groups = hamiltonian.group_up(lambda a, b: a.commute(b))
        
        self.stds = []
        self.change_of_basis_gates = []
        self.measurement_operators = []
        
        for group in self.groups:
            
            qv = QuantumVariable(n)
            
            meas_op = group.change_of_basis(qv, diagonalisation_method)
            self.change_of_basis_gates.append(qv.qs.to_gate())
            self.measurement_operators.append(meas_op)
            
            # Collect standard deviation
            self.stds.append(np.sqrt(meas_op.get_operator_variance(n = n)))
        
        N = sum(self.stds)
        self.shots_list = [N*s for s in self.stds]
    
    def get_measurement(self, qc, qubit_list, precision, backend):
        
        from qrisp.misc import get_measurement_from_qc
        results = []
        meas_ops = []
        meas_coeffs = []
        
        for i in range(len(self.measurement_operators)):
            
            group = self.measurement_operators[i]
            
            shots = int(self.shots_list[i]/precision**2)
            
            qubits = [qubit_list[j] for j in range(self.change_of_basis_gates[i].num_qubits)]
            
            curr = qc.copy()
            curr.append(self.change_of_basis_gates[i], qubits)
            
            res = get_measurement_from_qc(curr, list(qubit_list), backend, shots)
            results.append(res)
            
            temp_meas_ops = []
            temp_coeff = []
            for term, coeff in group.terms_dict.items():
                temp_meas_ops.append(term.serialize())
                temp_coeff.append(coeff)
                
            meas_coeffs.append(temp_coeff)
            meas_ops.append(temp_meas_ops)
        
        samples = create_padded_array([list(res.keys()) for res in results]).astype(np.int64)
        probs = create_padded_array([list(res.values()) for res in results])
        meas_ops = create_padded_array(meas_ops, use_tuples = True).astype(np.int64)
        meas_coeffs = create_padded_array(meas_coeffs)
        
        return evaluate_expectation_jitted(samples, probs, meas_ops, meas_coeffs)
        


def create_padded_array(list_of_lists, use_tuples = False):
    """
    Create a padded numpy array from a list of lists with varying lengths.
    
    Parameters:
    list_of_lists (list): A list of lists with potentially different lengths.
    
    Returns:
    numpy.ndarray: A 2D numpy array with padded rows.
    """
    # Find the maximum length of any list in the input
    max_length = max(len(lst) for lst in list_of_lists)
    
    # Create a padded list of lists
    if not use_tuples:
        padded_lists = [
            lst + [0] * (max_length - len(lst))
            for lst in list_of_lists
        ]
    else:
        padded_lists = [
            lst + [(0,0,0,0)] * (max_length - len(lst))
            for lst in list_of_lists
        ]
        
    
    # Convert to numpy array
    return np.array(padded_lists)


#
# Evaluate expectation
#


def evaluate_expectation(samples, probs, operators, coefficients):
    """
    Evaluate the expectation.
    
    """
    # print(results)
    # print(operators)
    # print(coefficients)
    # raise

    expectation = 0

    for index1,ops in enumerate(operators):
        for index2,op in enumerate(ops):
            for i in range(len(samples[index1])):
                outcome,probability = samples[index1, i], probs[index1, i]
                expectation += probability*evaluate_observable(op,outcome)*np.real(coefficients[index1][index2])
    
    return expectation


def evaluate_observable(observable: tuple, x: int):
    # This function evaluates how to compute the energy of a measurement sample x.
    # Since we are also considering ladder operators, this energy can either be
    # 0, -1 or 1. For more details check out the comments of QubitOperator.get_conjugation_circuit
    
    # The observable is given as tuple, containing for integers and a boolean.
    # To understand the meaning of these integers check QubitTerm.serialize.
    
    # Unwrap the tuple
    z_int, AND_bits, AND_ctrl_state, contains_ladder = observable

    # Compute whether the sign should be sign flipped based on the Z operators
    sign_flip_int = z_int & x
    sign_flip = 0
    while sign_flip_int:
        sign_flip += sign_flip_int & 1
        sign_flip_int >>= 1
    
    # If there is a ladder operator in the term, we need to half the energy 
    # because we want to measure (|110><110| - |111><111|)/2
    if contains_ladder:
        prefactor = 0.5
    else:
        prefactor = 1
    
    # If there are no and bits, we return the result
    if AND_bits == 0:
        return prefactor*(-1)**sign_flip

    # Otherwise we apply the AND_ctrl_state to flip the appropriate bits.
    corrected_x = (x ^ AND_ctrl_state)
    
    # If all bits are in the 0 state the AND is true.
    if corrected_x & AND_bits == 0:
        return prefactor*(-1)**sign_flip
    else:
        return 0
    

evaluate_observable_jitted = njit(cache = True)(evaluate_observable)

@njit(cache = True)
def evaluate_expectation_jitted(samples, probs, operators, coefficients):
    """
    Evaluate the expectation.
    
    """
    expectation = 0

    for index1,ops in enumerate(operators):
        for index2,op in enumerate(ops):
            for i in range(len(samples[index1])):
                outcome,probability = samples[index1, i], probs[index1, i]
                expectation += probability*evaluate_observable_jitted(op,outcome)*np.real(coefficients[index1][index2])
    
    return expectation



def partition(values, num_qubits):
    """
    Partitions a list of integers into a list of lists of integers with size 64 bit.

    Parameters
    ----------
    values : list[int]
        A list of integers.
    num_qubits : int
        The maximal number of bits.

    Returns
    -------
    partition : list[np.array]
        A list of NumPy numpy.uint64 arrays.

    """
    

    M = math.ceil(num_qubits/64)
    N = len(values)

    zeros = [[0]*N for k in range(M)]
    partition = [values]
    partition.extend(zeros)

    lower_mask = (1<<64) - 1
    
    for j in range(1,M):
        for i in range(N):
            partition[j][i] = partition[j-1][i] & lower_mask
            partition[j+1][i] = partition[j-1][i] >> 64

    if M==1:
        return [np.array(partition[0], dtype=np.uint64)]
    else:
        return [np.array(part, dtype=np.uint64) for part in partition[1:]]
    


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import time

import numpy as np
from scipy.optimize import minimize
from sympy import Symbol

from qrisp import QuantumArray
from qrisp.algorithms.vqe.vqe_benchmark_data import VQEBenchmark
from qrisp.operators.qubit.measurement import QubitOperatorMeasurement
from qrisp.operators.fermionic import FermionicOperator

class VQEProblem:
    r"""
    Central structure to facilitate treatment of VQE problems.
    This class encapsulates the Hamiltonian, the ansatz, and the initial state preparation function for a specific VQE problem instance. 
    
    Parameters
    ----------
    hamiltonian : :ref:`QubitOperator` or :ref:`FermionicOperator`
        The problem Hamiltonian.
    ansatz_function : function
        A function receiving a :ref:`QuantumVariable` or :ref:`QuantumArray` and a parameter list. This function implements the unitary 
        corresponding to one layer of the ansatz.
    num_params : int
        The number of parameters per layer.
    init_function : function, optional
        A function preparing the initial state.
        By default, the inital state is the $\ket{0}$ state.
    callback : bool, optional
        If ``True``, intermediate results are stored. The default is ``False``.

    Examples
    --------

    For a quick demonstration, we show how to calculate the ground state energy of the $H_2$ molecule using VQE, as explained `here <https://arxiv.org/abs/2305.07092>`_.

    ::

        from qrisp import *
        from qrisp.operators.qubit import X,Y,Z

        # Problem Hamiltonian
        c = [-0.81054, 0.16614, 0.16892, 0.17218, -0.22573, 0.12091, 0.166145, 0.04523]
        H = c[0] \
            + c[1]*Z(0)*Z(2) \
            + c[2]*Z(1)*Z(3) \
            + c[3]*(Z(3) + Z(1)) \
            + c[4]*(Z(2) + Z(0)) \
            + c[5]*(Z(2)*Z(3) + Z(0)*Z(1)) \
            + c[6]*(Z(0)*Z(3) + Z(1)*Z(2)) \
            + c[7]*(Y(0)*Y(1)*Y(2)*Y(3) + X(0)*X(1)*Y(2)*Y(3) + Y(0)*Y(1)*X(2)*X(3) + X(0)*X(1)*X(2)*X(3))

        # Ansatz
        def ansatz(qv,theta):
            for i in range(4):
                ry(theta[i],qv[i])
            for i in range(3):
                cx(qv[i],qv[i+1])
            cx(qv[3],qv[0])

        from qrisp.vqe.vqe_problem import *

        vqe = VQEProblem(hamiltonian = H,
                         ansatz_function = ansatz,
                         num_params=4,
                         callback=True)

        energy = vqe.run(qarg = QuantumVariable(4),
                      depth = 1,
                      max_iter=50)
        print(energy)
        # Yields -1.864179046
    
    Note that for comparing to the results in the aforementioned paper, we have to add the nuclear repulsion energy $E_{\text{nuc}}=0.72$ to the calculated electronic energy $E_{\text{el}}$.

    We visualize the optimization process:

    >>> vqe.visualize_energy(exact=True)

    .. figure:: /_static/vqeH2.png
        :alt: VQEH2
        :scale: 80%
        :align: center

    """
    
    def __init__(self, hamiltonian, ansatz_function, num_params, init_function = None, callback=False):
        
        if isinstance(hamiltonian, FermionicOperator):
            hamiltonian = hamiltonian.to_qubit_operator()
            
        self.hamiltonian = hamiltonian
        self.ansatz_function = ansatz_function
        self.num_params = num_params
        self.init_function = init_function
        self.cl_post_processor = None

        # parameters for callback
        self.callback = callback
        self.optimization_params = []
        self.optimization_costs = []

    def set_callback(self):
        """
        Sets ``callback=True`` for saving intermediate results.

        """

        self.callback = True

    def set_init_function(self, init_function):
        """
        Set the initial state preparation function for the VQE problem.

        Parameters
        ----------
        init_function : function
            The initial state preparation function for the specific VQE problem instance.
        """
        self.init_function = init_function
      
    def compile_circuit(self, qarg, depth):
        """
        Compiles the circuit that is evaluated by the :meth:`run <qrisp.vqe.VQEProblem.run>` method.

        Parameters
        ----------
        qarg : :ref:`QuantumVariable` or :ref:`QuantumArray`
            The argument to which the VQE circuit is applied.
        depth : int
            The amount of VQE layers.

        Returns
        -------
        compiled_qc : QuantumCircuit
            The parametrized, compiled QuantumCircuit without measurements.
        list[sympy.Symbol]
            A list of the parameters that appear in ``compiled_qc``.

        """
        
        temp = list(qarg.qs.data)
        
        # Define parameters theta for VQE circuit
        theta = [Symbol("theta_" + str(i) + str(j)) for i in range(depth) for j in range(self.num_params)]

        # Prepare the initial state for particular problem instance, the default is the \ket{0} state
        if self.init_function is not None:
            self.init_function(qarg)
            
        # Apply p layers of the ansatz
        for i in range(depth):                           
            self.ansatz_function(qarg,[theta[self.num_params*i+j] for j in range(self.num_params)])

        # Compile quantum circuit
        compiled_qc = qarg.qs.compile()
        
        qarg.qs.data = temp
        
        return compiled_qc, theta

    def optimization_routine(self, qarg, depth, mes_kwargs, max_iter, init_type="random", init_point=None, measurement_data=None, optimizer="COBYLA"): 
        """
        Wrapper subroutine for the optimization method used in QAOA. The initial values are set and the optimization via ``COBYLA`` is conducted here.

        Parameters
        ----------
        qarg : :ref:`QuantumVariable` or :ref:`QuantumArray`
            The argument the cost function is called on.
        depth : int
            The amont of VQE layers.
        mes_kwargs : dict
            The keyword arguments for the measurement function.
        max_iter : int
            The maximum number of iterations for the optimization method.
        init_type : string, optional
            Specifies the way the initial optimization parameters are chosen. Available is ``random``. 
            The default is ``random``: Parameters are initialized uniformly at random in the interval $[0,\pi/2)]$.
        init_point : list[float], optional
            Specifies the initial optimization parameters. 
        measurement_data : QubitOperatorMeasurement
            Cached data to accelerate the measurement procedure. Automatically generated by default.
        optimizer : str, optional
            Specifies the optimization routine. Available are ``COBYLA``, ``COBYQA``, ``Nelder-Mead``. 
            The Default is "COBYLA". 

        Returns
        -------
        res_sample
            The optimized parameters of the problem instance.
        """

        # Define optimization wrapper function to be minimized using VQE
        def optimization_wrapper(theta, qc, symbols, qarg, measurement_data, mes_kwargs):
            """
            Wrapper function for the optimization method used in VQE.

            This function calculates the expected value of the Hamiltonian after post-processing if a post-processing function is set, otherwise it calculates the expected value of the Hamiltonian.

            Parameters
            ----------
            theta : list
                The list of angle parameters for the VQE circuit.
            qc : QuantumCircuit
                The compiled quantum circuit.
            symbols : list
                The list of symbols used in the quantum circuit.
            qarg_dupl : :ref:`QuantumVariable`
                The duplicated quantum variable to which the quantum circuit is applied.
            mes_kwargs : dict
                The keyword arguments for the measurement function.

            Returns
            -------
            float
                The expected value of the Hamiltonian.
            """         

            subs_dic = {symbols[i] : theta[i] for i in range(len(symbols))}

            expectation = self.hamiltonian.get_measurement(qarg, 
                                                           subs_dic = subs_dic, 
                                                            measurement_data = measurement_data, 
                                                           precompiled_qc = qc, **mes_kwargs)
            

            if self.callback:
                self.optimization_costs.append(expectation)

            if self.cl_post_processor is not None:
                return self.cl_post_processor(expectation)
            else:
                return expectation

        if init_point is None:
            # Set initial random values for optimization parameters 
            if init_type=='random':
                init_point = np.pi * np.random.rand(depth * self.num_params)/2

        #def optimization_cb(x):
        #    self.optimization_params.append(x)

        # Perform optimization using specified method
        compiled_qc, symbols = self.compile_circuit(qarg, depth)
        res_sample = minimize(optimization_wrapper,
                                init_point, 
                                method=optimizer,
                                options={'maxiter':max_iter}, 
                                args = (compiled_qc, symbols, qarg, measurement_data, mes_kwargs))
            
        return res_sample['x']

    def run(self, qarg, depth, mes_kwargs = {}, max_iter = 50, init_type = "random", init_point=None, optimizer="COBYLA"):
        """
        Run the specific VQE problem instance with given quantum arguments, depth of VQE circuit,
        measurement keyword arguments (mes_kwargs) and maximum iterations for optimization (max_iter).
        
        Parameters
        ----------
        qarg : :ref:`QuantumVariable` or :ref:`QuantumArray`
            The argument to which the VQE circuit is applied.
        depth : int
            The amount of VQE layers.
        mes_kwargs : dict, optional
            The keyword arguments for the measurement function. Default is an empty dictionary.
            By default, the target ``precision`` is set to 0.01, and the maximum amount of ``shots`` is 100000.
        max_iter : int, optional
            The maximum number of iterations for the optimization method. Default is 50.
        init_type : string, optional
            Specifies the way the initial optimization parameters are chosen. Available is ``random``. 
            The default is ``random``: Parameters are initialized uniformly at random in the interval $[0,\pi/2)]$.
        init_point : list[float], optional
            Specifies the initial optimization parameters. 
        optimizer : str, optional
            Specifies the `optimization routine <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_. 
            Available are, e.g., ``COBYLA``, ``COBYQA``, ``Nelder-Mead``. The Default is ``COBYLA``. 

        Returns
        -------
        energy : float
            The expected value of the Hamiltonian after applying the optimal VQE circuit to the quantum argument.
        """

        # Delete callback
        self.optimization_params = []
        self.optimization_costs = []

        if not "shots" in mes_kwargs:
            mes_kwargs["shots"] = 1000000
        if not "precision" in mes_kwargs:
            mes_kwargs["precision"] = 0.01

        measurement_data = QubitOperatorMeasurement(self.hamiltonian)
        
        optimal_theta = self.optimization_routine(qarg, 
                                                  depth, 
                                                  mes_kwargs, 
                                                  max_iter, 
                                                  init_type, 
                                                  init_point, 
                                                  measurement_data=measurement_data, 
                                                  optimizer = optimizer)
        
        # Prepare the initial state for particular problem instance, the default is the \ket{0} state
        if self.init_function is not None:
            self.init_function(qarg)

        # Apply p layers of the ansatz    
        for i in range(depth):                          
            self.ansatz_function(qarg,[optimal_theta[self.num_params*i+j] for j in range(self.num_params)])

        opt_res = self.hamiltonian.get_measurement(qarg,**mes_kwargs, measurement_data=measurement_data)
        
        return opt_res
    
    def train_function(self, qarg, depth, mes_kwargs = {}, max_iter = 50, init_type = "random", init_point=None, optimizer="COBYLA"):
        """
        This function allows for training of a circuit with a given instance of a ``VQEProblem``. It will then return a function that can be applied to a :ref:`QuantumVariable`,
        such that it prepares the ground state of the problem Hamiltonian. The function therefore applies a circuit for the problem instance with optimized parameters.

        Parameters
        ----------
        qarg : :ref:`QuantumVariable` 
            The argument to which the VQE circuit is applied.
        depth : int
            The amount of VQE layers.
        mes_kwargs : dict, optional
            The keyword arguments for the measurement function. Default is an empty dictionary.
            By default, the target ``precision`` is set to 0.01, and the maximum amount of ``shots`` is 100000.
        max_iter : int, optional
            The maximum number of iterations for the optimization method. Default is 50.
        init_type : string, optional
            Specifies the way the initial optimization parameters are chosen. Available is ``random``. 
            The default is ``random``: Parameters are initialized uniformly at random in the interval $[0,\pi/2)]$.
        init_point : list[float], optional
            Specifies the initial optimization parameters. 
        optimizer : str, optional
            Specifies the `optimization routine <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_. 
            Available are, e.g., ``COBYLA``, ``COBYQA``, ``Nelder-Mead``. The Default is "COBYLA". 

        Returns
        -------
        circuit_generator : function
            A function that can be applied to a :ref:`QuantumVariable`, with optimized parameters for the problem instance. 
            The :ref:`QuantumVariable` then represents the ground state of the problem Hamiltonian.

        """

        if not "shots" in mes_kwargs:
            mes_kwargs["shots"] = 1000000
        if not "precision" in mes_kwargs:
            mes_kwargs["precision"] = 0.01

        measurement_data = QubitOperatorMeasurement(self.hamiltonian)

        optimal_theta = self.optimization_routine(qarg, depth, mes_kwargs, max_iter, init_type, init_point, optimizer, measurement_data=measurement_data)
        
        def circuit_generator(qarg_gen):
            # Prepare the initial state for particular problem instance, the default is the \ket{0} state
            if self.init_function is not None:
                self.init_function(qarg_gen)

            for i in range(depth): 
                self.ansatz_function(qarg_gen,[optimal_theta[self.num_params*i+j] for j in range(self.num_params)])
            
        return circuit_generator
    
    def benchmark(self, qarg, depth_range, shot_range, iter_range, optimal_energy, repetitions = 1, mes_kwargs = {}, init_type = "random"):
        """
        This method enables convenient data collection regarding performance of the implementation.

        Parameters
        ----------
        qarg : :ref:`QuantumVariable` or :ref:`QuantumArray`
            The quantum argument the benchmark is executed on. Compare to the :meth:`.run <qrisp.vqe.VQEProblem.run>` method.
        depth_range : list[int]
            A list of integers indicating, which depth parameters should be explored. Depth means the amount of VQE layers.
        shot_range : list[int]
            A list of integers indicating, which shots parameters should be explored. Shots means the amount of repetitions, the backend performs per iteration and per measurement setting.
        iter_range : list[int]
            A list of integers indicating, what iterations parameter should be explored. Iterations means the amount of backend calls, the optimizer is allowed to do.
        optimal_energy: float
            The exact ground state energy of the problem Hamiltonian. 
        repetitions : int, optional
            The amount of repetitions, each parameter constellation should go though. Can be used to get a better statistical significance. The default is 1.
        mes_kwargs : dict, optional
            The keyword arguments, that are used for the ``qarg.get_spin_measurement``. The default is {}.
        init_type : string, optional
            Specifies the way the initial optimization parameters are chosen. Available is ``random``. 
            The default is ``random``: Parameters are initialized uniformly at random in the interval $[0,\pi/2)]$.

        Returns
        -------
        :ref:`VQEBenchmark`
            The results of the benchmark.
            
        Examples
        --------
        
        We create a Heisenberg problem instance and benchmark several parameters:
        
        ::
            
            from networkx import Graph

            G =Graph()
            G.add_edges_from([(0,1),(1,2),(2,3),(3,4)])
            from qrisp.vqe.problems.heisenberg import *

            vqe = heisenberg_problem(G,1,0)
            H = create_heisenberg_hamiltonian(G,1,0)

            benchmark_data = vqe.benchmark(qarg = QuantumVariable(5),
                                depth_range = [1,2,3],
                                shot_range = [5000,10000],
                                iter_range = [25,50],
                                optimal_energy = H.ground_state_energy(),
                                repetitions = 2
                                )
        
        We can investigate the data by calling ``visualize``:
        
        ::
            
            benchmark_data.visualize()
        
        .. image:: vqe_benchmark_plot.png
            
        The :ref:`VQEBenchmark` class contains a variety of methods to help 
        you drawing conclusions from the collected data. Make sure to check them out!

        """
        
        data_dict = {"layer_depth" : [],
                     "circuit_depth" : [],
                     "qubit_amount" : [],
                     "shots" : [],
                     "iterations" : [],
                     "runtime" : [],
                     "energy" : []
                     }
        
        for p in depth_range:
            for s in shot_range:
                for it in iter_range:
                    for k in range(repetitions):
                        
                        if isinstance(qarg, QuantumArray):
                            qarg_dupl = QuantumArray(qtype = qarg.qtype, shape = qarg.shape)
                        else:
                            qarg_dupl = qarg.duplicate()
                            
                        start_time = time.time()
                        
                        temp_mes_kwargs = dict(mes_kwargs)
                        temp_mes_kwargs["shots"] = s

                        energy = self.run(qarg=qarg_dupl, depth = p, max_iter = it, mes_kwargs = temp_mes_kwargs, init_type=init_type)

                        final_time = time.time() - start_time
                    
                        compiled_qc = qarg_dupl.qs.compile()
                        
                        data_dict["layer_depth"].append(p)
                        data_dict["circuit_depth"].append(compiled_qc.depth())
                        data_dict["qubit_amount"].append(compiled_qc.num_qubits())
                        data_dict["shots"].append(s)
                        data_dict["iterations"].append(it)
                        data_dict["energy"].append(energy)
                        data_dict["runtime"].append(final_time)
                        
                        
        return VQEBenchmark(data_dict, optimal_energy, self.hamiltonian)
    

    def visualize_energy(self,exact=False):
        """
        Visualizes the energy during the optimization process.

        Parameters
        ----------
        exact : Boolean
            If ``True``, the exact ground state energy of the spin operator is computed classically, and compared to the energy in the optimization process.
            The default is ``False``.

        """

        import matplotlib.pyplot as plt

        if not self.callback:
            raise Exception("Visualization can only be performed for a VQE instance with callback=True")
        
        if exact:
            exact_solution = self.hamiltonian.ground_state_energy()
            plt.axhline(y=exact_solution, color="#6929C4", linestyle='--', linewidth=2, label='Exact energy')
        
        x = list(range(len(self.optimization_costs)))
        y = self.optimization_costs
        plt.scatter(x, y, color='#20306f',marker="o", linestyle='solid', linewidth=1, label='VQE energy')
        plt.xlabel("Iterations", fontsize=15, color="#444444")
        plt.ylabel("Energy", fontsize=15, color="#444444")
        plt.legend(fontsize=12, labelcolor="#444444")
        plt.tick_params(axis='both', labelsize=12)
        plt.grid()

        plt.show()



import math 
from qrisp import cx, RYGate, ry, x, rzz, rz
import numpy as np 



def portfolio_cost_operator(problem):
    """
    | Quantum cost operator  for the discrete portfolio rebalancing problem, as described in https://arxiv.org/pdf/1911.05296.pdf.
    | It is depended on the problem instance, including the old portfolio positions, the normalized covariance matrix, the normalized asset returns and trading costs. See example implementation for formatting. 

    Parameters
    ----------
    problem : List
        A list containing a the relevant data for the problem instance.

    Returns
    -------
    portfolio_cost_op: function
        A callable function to be applied to a QuantumArray for solving the problem instance.

    """


    old_pos = problem[0]
    risk_return = problem[1] 
    covar_matrix = problem[2]
    asset_return = problem[3] 
    #tradiing_cost
    tc = problem[4]/4
    def portfolio_cost_op(q_array, gamma):
        #does this work???? -- seems like it would
        #print("cost_op")
        #print(len(qv))
        l = q_array[1]
        s = q_array[0]

        #risk-return function
        for i in range(len(s)):
            prefac = 0
            for j in range(len(s)):
                
                prefac = gamma * covar_matrix[i][j]/4 * risk_return
                rzz(-prefac, l[i], s[j])
                rzz(-prefac, s[i], l[j])

                if i == j:
                    # case of rzz on same spin vars
                    rz(2*prefac, l[i])
                    rz(2*prefac, s[i])
                    continue

                rzz(prefac, l[i], l[j])
                rzz(prefac, s[i], s[j])
            
            prefac2 = gamma *asset_return[i] * (1-risk_return)/2
            rz(prefac2, l[i])
            rz(- prefac2, s[i])

        #trading cost function
        # first added term can be dropped, as it just results in a phase
        for i in range(len(s)):
            old_val = old_pos[i]
            rz(gamma* tc*(1 - old_val^2 - old_val), l[i])
            rz(gamma* tc*(1 - old_val^2 - old_val), s[i])
            rzz(gamma* tc* (2*old_val^2 -1), l[i], s[i] )

    return portfolio_cost_op


def portfolio_cl_cost_function(problem):
    """
    | Classical cost function for the discrete portfolio rebalancing problem, as described in https://arxiv.org/pdf/1911.05296.pdf.
    |It is depended on the problem instance, including the old portfolio positions, the normalized covariance matrix, the normalized asset returns and trading costs. See example implementation for formatting. 

    Parameters
    ----------
    problem : list
        A list containing a the relevant data for the problem instance.

    Returns
    -------
    cl_cost_function : function
        A callable function to calculate the cost value of the problem solution.

    """

    old_pos = problem[0]
    risk_return = problem[1] 
    covar_matrix = problem[2]
    asset_return = problem[3] 
    #tradiing_cost
    tc = problem[4]
    def cl_cost_function(res):
        #print(res)
        energy = 0
        counts = 0
        half = int(len(list(res.keys())[0][0]))
        
        key_list = []
        for key, val in res.items():
            
            #half = len(key)/2
            #new_key = [int(key[i])-int(key[i+half]) for i in range(half)]
            new_key = [int(key[0][i])-int(key[1][i]) for i in range(half)] # ??????
            key_list.append(new_key)
            rr1 = sum([risk_return*covar_matrix[i][j] *new_key[i]*new_key[j] for i in range(half) for j in range(half)])
            rr2 = sum([(1-risk_return)*asset_return[j] *new_key[j] for j in range(half)])
            c_tc= sum([tc  for i in range(half) if new_key[i] != old_pos[i]])
            energy -= (rr1+ rr2+ c_tc)*val
            counts += val
        final = energy/counts 
        #print(final)  
        return final
    
    return cl_cost_function





def portfolio_init(lots):
    """
    | Initial state for the discrete portfolio rebalancing problem, as described in https://arxiv.org/abs/1904.07358.
    | Depending on the number of lots a QuantumArray is prepared, where the first index describes the short positions held and the second index describes the long postions held.

    Parameters
    ----------
    lots : Int
        The number of lots in the initial portfolio position

    Returns
    -------
    state_prep: function
        A callable function to be applied to a QuantumArray to receive the initial state for the problem.

    """

    def state_prep(q_array):

        l = q_array[1]
        s = q_array[0]
        
        n = len(l)
        band_prefix = dict()
        max_pref = 0
        for index in range(n- lots +1): 
            max_pref += math.comb(n, index)*math.comb(n, lots+ index)
            this_pref = math.comb(n, index)*math.comb(n, lots+ index)
            band_prefix.setdefault(str(index), this_pref)

        x(l[-lots:])
        param = 2 * np.arccos(np.sqrt((band_prefix["0"])/(max_pref)))
        ry(param,  s[-1])
        qc_s = s[-1].qs()

        # how does one do the superpos? is everything controlled with everything that came before? or just the first? or only one before?
        for index1 in range(1,lots):
            param = 2 * np.arccos(np.sqrt((band_prefix[str(index1)])/(max_pref)))
            cry_gate = RYGate(param).control(1)
            qc_s.append(cry_gate, [s[-index1], s[-index1-1]])

        # the lots+ index thing below may cause problems in the future...
        for index2 in range(1,lots+1):
            cx(s[-index2],l[-lots -index2])
    return state_prep



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from flask import Flask, request, jsonify, make_response
import yaml
from datetime import date, datetime
import threading

"""

This file sets up a server adhering to the interface specified by the Qunicorn middleware
developed in the SeQuenC project: https://sequenc.de/

If you are a backend provider, we don't recommend using this server for production,
since it only exposes the necessities to run quantum circuits and nothing else.
Instead check out the Qunicorn GitHub: https://github.com/qunicorn/qunicorn-core
And it's documentation on how to create new Pilots:
https://qunicorn-core.readthedocs.io/en/latest/tutorials/pilot_tutorial_demo.html

"""

# Returns the hosts ip
def get_ip():
    import socket

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(("10.255.255.255", 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = "127.0.0.1"
    finally:
        s.close()
    return IP


class BackendServer:
    """
    This class allows convenient setup of a server respecting the `Qunicorn <https://qunicorn-core.readthedocs.io/en/latest/index.html>`_ interface.

    Parameters
    ----------
    run_func : function
        A function that receives a QuantumCircuit, an integer specifying the amount of
        shots and a token in the form of a string. It returns the counts as a dictionary
        of bitstrings.
    ip_address : str, optional
        The IP address of where the listening socket should be opened. By default, the
        IP address of the hosting machine will be used.
    port : int, optional
        The port on which to listen for requests. By default, 9010 will be used.
    is_simulator : bool, optional
        A bool specifying whether the results returned by this server are due to a
        simulator. The default is True.

    Examples
    --------

    We create a server listening on the localhost IP address using a run function which
    prints the token and queries the Aer-simulator. ::

        def run_func(qasm_str, shots):
            
            from qiskit import QuantumCircuit
            #Convert to qiskit
            
            qiskit_qc = QuantumCircuit.from_qasm_str(qasm_str)

            from qiskit_aer import AerSimulator
            qiskit_backend = AerSimulator()

            #Run Circuit on the Qiskit backend
            return qiskit_backend.run(qiskit_qc, shots = shots).result().get_counts()

        from qrisp.interface import BackendServer
        example_server = BackendServer(run_func, ip_address = "127.0.0.1", port = 8080)
        example_server.start()

    """
    
    def __init__(self, run_func , ip_address = None, port = None):

        self.app = Flask(__name__)
        self.deployments = []
        if ip_address is None:
            ip_address = get_ip()
        if port is None:
            port = 9010
        self.port = port
        self.ip_address = ip_address 
        self.jobs = []
        self.run_func = run_func
        self.is_simulator = True
        # To set up the server we route the relevant functions for the flask app.
        
        #This function receives the deployment data and saves it in the 
        #deployment attribute
        @self.app.route('/deployments', methods=['POST'])
        def deployments():
            
            # Receive the quantum circuit data from the request
            data = request.get_json()
            
            #Append the deployment data to the library
            self.deployments.append(data)
                
            #For the response dictionary
            response = {
                      "id": len(self.deployments)-1,
                      "deployedBy": {
                        "id": 0,
                        "name": ""
                      },
                      "deployedAt": str(date.today()),
                      "name": ""
                    }
            
            response["programs"] = data["programs"]
            
            #Create response object
            response_object = make_response(jsonify(response), 201) 
            
            return response_object
        
        #This function receives the deployment id of a job and starts the execution
        @self.app.route('/jobs', methods=['POST'])
        def jobs_post():
            
            # Receive the quantum circuit data from the request
            data = request.get_json()
            
            #Extract the deployment
            deployment_id = data["deploymentId"]
            deployment = self.deployments[deployment_id]
            
            #Create the wrapper for asynchronous execution of the job
            def job_wrapper():
                for i in range(len(deployment["programs"])):
                    counts_res = self.run_func(deployment["programs"][i]["quantumCircuit"], 
                                                data["shots"], 
                                                data["token"])
                    job_dic["counts"].append(counts_res)
                # try:
                #     for i in range(len(deployment["programs"])):
                #         counts_res = self.run_func(deployment["programs"][i]["quantumCircuit"], 
                #                                     data["shots"], 
                #                                     data["token"])
                #         job_dic["counts"].append(counts_res)
                # except Exception as e:
                #     job_dic["exception"] = e
                    
                job_dic["end_time"] = str(datetime.now())
            
            #Create the thread
            run_thread = threading.Thread(target = job_wrapper)    
            
            #Create the dictionary to hold information about the job 
            job_dic = {}
            
            job_dic["deploymentId"] = deployment_id
            job_dic["counts"] = []
            job_dic["shots"] = data["shots"]
            job_dic["exception"] = None
            job_dic["quantumCircuits"] = deployment["programs"]
            job_dic["run_thread"] = run_thread
            job_dic["start_time"] = str(datetime.now())
            job_dic["end_time"] = "-"
            
            
            #Start the thread            
            run_thread.start()
            
            #Append the job to jobs attribute            
            self.jobs.append(job_dic)
            
            #Form the response
            response = {
                    "id": len(self.jobs)-1,
                    "jobName": "",
                    "jobState": "running"
                    }
            response_object = make_response(jsonify(response), 201) 
            
            return response_object
        
        #This function retrieves information about a certain job. It is used by
        #the client to inquire the results of the job
        @self.app.route('/jobs/<int:job_id>/', methods=['GET'])
        def get_job(job_id):
            
            # Search for the job with the specified job_id
            if not job_id < len(self.jobs):
                return jsonify({"error": "Job not found"}), 404
            
            #Get job information
            job = self.jobs[job_id]
            
            #Find out wether the job finished
            thread = job["run_thread"]
            
            if not thread.is_alive():
                state = "finished"
            else:
                state = "running"
            
            #Find out wether the job suceeded
            if job["exception"] is not None:
                
                state = "failed"
                response = {
                  "code": 0,
                  "status": "failed",
                  "message": str(job["exception"]),
                  "errors": {}
                }
                
                response_object = make_response(jsonify(response)) 
                return response_object
            
            #Form the response
            response = {
              "id": job_id,
              "executedBy": {
                "id": 0,
                "name": ""
              },
              "executedOn": {
                "id": 0,
                "numQubits": -1,
                "isSimulator": self.is_simulator,
                "isLocal": True,
                "provider": {
                  "id": 0,
                  "withToken": True,
                  "supportedLanguages": {
                    "id": 0,
                    "providerId": "string",
                    "name": ""
                  },
                  "name": ""
                }
              },
              "progress": 0,
              "state": state,
              "type": "",
              "startedAt": job["start_time"],
              "finishedAt": job["end_time"],
              "data": "",
              "results": [{
                "id": 0,
                "circuit": job["quantumCircuits"][i],
                "results": job["counts"][i],
                "resultType": "COUNTS",
                "metaData": {}
              } for i in range(len(job["counts"]))],
              "parameters": ""
            }
    
            response_object = make_response(jsonify(response), 201)
            
            return response_object

    def start(self):
        
        from waitress import serve
        
        #Set up the thread to run the server    
        def wrapper():
            serve(self.app, host=self.ip_address, port=self.port)
            # self.app.run(host=self.ip_address, port=self.port)
        thr = threading.Thread(target=wrapper)
        thr.setDaemon(True)

        # Start the thread
        thr.start()
        
        import requests
        # Hold programm until the server answers
        while True:
            try:
                response = requests.get("http://" + self.ip_address + ":" + str(self.port) + "/jobs/0/")
            except:
                continue
            break
            
            



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

def amplitude_amplification(args, state_function, oracle_function, kwargs_oracle={}, iter=1):
    r"""
    This method performs `quantum amplitude amplification <https://arxiv.org/abs/quant-ph/0005055>`_.

    The problem of quantum amplitude amplification is described as follows:

    * Given a unitary operator :math:`\mathcal{A}`, let :math:`\ket{\Psi}=\mathcal{A}\ket{0}`.
    * Write :math:`\ket{\Psi}=\ket{\Psi_1}+\ket{\Psi_0}` as a superposition of the orthogonal good and bad components of :math:`\ket{\Psi}`.
    * Enhance the probability :math:`a=\langle\Psi_1|\Psi_1\rangle` that a measurement of $\ket{\Psi}$ yields a good state.

    Let $\theta_a\in [0,\pi/2]$ such that $\sin^2(\theta_a)=a$. Then the amplitude amplification operator $\mathcal Q$ acts as

    .. math::

        \mathcal Q^j\ket{\Psi}=\frac{1}{\sqrt{a}}\sin((2j+1)\theta_a)\ket{\Psi_1}+\frac{1}{\sqrt{1-a}}\cos((2j+1)\theta_a)\ket{\Psi_0}.

    Therefore, after $m$ iterations the probability of measuring a good state is $\sin^2((2m+1)\theta_a)$. 
    
    Parameters
    ----------

    args : QuantumVariable or list[QuantumVariable]
        The (list of) QuantumVariables which represent the state,
        the amplitude amplification is performed on.
    state_function : function
        A Python function preparing the state $\ket{\Psi}$.
        This function will receive the variables in the list ``args`` as arguments in the
        course of this algorithm.
    oracle_function : function
        A Python function tagging the good state $\ket{\Psi_1}$.
        This function will receive the variables in the list ``args`` as arguments in the
        course of this algorithm.
    kwargs_oracle : dict, optional
        A dictionary containing keyword arguments for the oracle. The default is {}.
    iter : int, optional
        The amount of amplitude amplification iterations to perform. The default is 1.

    Examples
    --------

    We define a function that prepares the state :math:`\ket{\Psi}=\cos(\frac{\pi}{16})\ket{0}+\sin(\frac{\pi}{16})\ket{1}`
    and an oracle that tags the good state :math:`\ket{1}`. In this case, we have :math:`a=\sin^2(\frac{\pi}{16})\approx 0.19509`.
     
    ::

        from qrisp import z, ry, QuantumBool, amplitude_amplification
        import numpy as np

        def state_function(qb):
            ry(np.pi/8,qb)

        def oracle_function(qb):   
            z(qb)
        
        qb = QuantumBool()

        state_function(qb)

    >>> qb.qs.statevector(decimals=5)
    0.98079∣False⟩+0.19509∣True⟩

    We can enhance the probability of measuring the good state with amplitude amplification:

    >>> amplitude_amplification([qb], state_function, oracle_function)
    >>> qb.qs.statevector(decimals=5)
    0.83147*|False> + 0.55557*|True> 

    >>> amplitude_amplification([qb], state_function, oracle_function)
    >>> qb.qs.statevector(decimals=5)
    0.55557*|False> + 0.83147*|True> 

    >>> amplitude_amplification([qb], state_function, oracle_function)
    >>> qb.qs.statevector(decimals=5)
    0.19509*|False> + 0.98079*|True>

    """

    from qrisp import merge, IterationEnvironment, recursive_qs_search
    from qrisp.grover import diffuser

    merge(args)
    qs = recursive_qs_search(args)[0]
    with IterationEnvironment(qs, iter):
        oracle_function(*args, **kwargs_oracle)
        diffuser(args, state_function=state_function)



"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""
from qrisp.operators.hamiltonian import Hamiltonian
from qrisp.operators.qubit.bound_qubit_term import BoundQubitTerm
from qrisp.operators.qubit.measurement import get_measurement
from qrisp import h, sx, IterationEnvironment, conjugate, merge

import sympy as sp

threshold = 1e-9

#
# BoundQubitOperator
#

class BoundQubitOperator(Hamiltonian):
    r"""
    This class provides an efficient implementation of Pauli Hamiltonians acting on QuantumVariables, i.e.,
    Hamiltonians of the form

    .. math::
        
        H=\sum\limits_{j}\alpha_jP_j 
            
    where $P_j=\prod_i\sigma_i^j$ is a Pauli product, 
    and $\sigma_i^j\in\{I,X,Y,Z\}$ is the Pauli operator acting on qubit $i$.

    Parameters
    ----------
    terms_dict : dict, optional
        A dictionary representing a BoundQubitOperator.

    Examples
    --------

    A BoundQubitOperator can be specified conveniently in terms of ``X``, ``Y``, ``Z`` operators:

    ::

        from qrisp import QuantumVariable
        from qrisp.operators import BoundQubitOperator, X,Y,Z
        
        qv = QuantumVariable(2)
        H = 1+2*X(qv[0])+3*X(qv[0])*Y(qv[1])

    Yields $1+2X(qv.0)+3X(qv.0)Y(qv.1)$.

    """

    def __init__(self, terms_dict={}):
        self.terms_dict = terms_dict
        pass

    def len(self):
        return len(self.terms_dict)
    
    #
    # Printing
    #

    def _repr_latex_(self):
        # Convert the sympy expression to LaTeX and return it
        expr = self.to_expr()
        return f"${sp.latex(expr)}$"
    
    def __str__(self):
        # Convert the sympy expression to a string and return it
        expr = self.to_expr()
        return str(expr)
    
    def to_expr(self):
        """
        Returns a SymPy expression representing the operator.

        Returns
        -------
        expr : sympy.expr
            A SymPy expression representing the operator.

        """
        
        expr = 0  
        for pauli,coeff in self.terms_dict.items():
            expr += coeff*pauli.to_expr()
        return expr

    #
    # Arithmetic
    #

    def __pow__(self, e):
        if self.len()==1:
            if isinstance(e, int) and e>=0:
                if e%2==0:
                    return BoundQubitOperator({BoundQubitTerm():1})
                else:
                    return self
            else:
                raise TypeError("Unsupported operand type(s) for ** or pow(): "+str(type(self))+" and "+str(type(e)))
        else:
            raise TypeError("Unsupported operand type(s) for ** or pow(): "+str(type(self))+" and "+str(type(e)))

    def __add__(self,other):
        """
        Returns the sum of the operator self and other.

        Parameters
        ----------
        other : int, float, complex or BoundQubitOperator
            A scalar or a BoundQubitOperator to add to the operator self.

        Returns
        -------
        result : BoundQubitOperator
            The sum of the operator self and other.

        """

        if isinstance(other,(int,float,complex)):
            other = BoundQubitOperator({BoundQubitTerm():other})
        if not isinstance(other,BoundQubitOperator):
            raise TypeError("Cannot add BoundQubitOperator and "+str(type(other)))

        res_terms_dict = {}

        for pauli,coeff in self.terms_dict.items():
            res_terms_dict[pauli] = res_terms_dict.get(pauli,0)+coeff
            if abs(res_terms_dict[pauli])<threshold:
                del res_terms_dict[pauli]
    
        for pauli,coeff in other.terms_dict.items():
            res_terms_dict[pauli] = res_terms_dict.get(pauli,0)+coeff
            if abs(res_terms_dict[pauli])<threshold:
                del res_terms_dict[pauli]
        
        result = BoundQubitOperator(res_terms_dict)
        return result
    
    def __sub__(self,other):
        """
        Returns the difference of the operator self and other.

        Parameters
        ----------
        other : int, float, complex or BoundQubitOperator
            A scalar or a BoundQubitOperator to substract from the operator self.

        Returns
        -------
        result : BoundQubitOperator
            The difference of the operator self and other.

        """

        if isinstance(other,(int,float,complex)):
            other = BoundQubitOperator({BoundQubitTerm():other})
        if not isinstance(other,BoundQubitOperator):
            raise TypeError("Cannot substract BoundQubitOperator and "+str(type(other)))

        res_terms_dict = {}

        for pauli,coeff in self.terms_dict.items():
            res_terms_dict[pauli] = res_terms_dict.get(pauli,0)+coeff
            if abs(res_terms_dict[pauli])<threshold:
                del res_terms_dict[pauli]
    
        for pauli,coeff in other.terms_dict.items():
            res_terms_dict[pauli] = res_terms_dict.get(pauli,0)-coeff
            if abs(res_terms_dict[pauli])<threshold:
                del res_terms_dict[pauli]
        
        result = BoundQubitOperator(res_terms_dict)
        return result

    def __rsub__(self,other):
        """
        Returns the difference of the operator other and self.

        Parameters
        ----------
        other : int, float, complex or BoundQubitOperator
            A scalar or a BoundQubitOperator to substract the operator self from.

        Returns
        -------
        result : BoundQubitOperator
            The difference of the operator other and self.

        """

        if isinstance(other,(int,float,complex)):
            other = BoundQubitOperator({BoundQubitTerm():other})
        if not isinstance(other,BoundQubitOperator):
            raise TypeError("Cannot substract BoundQubitOperator and "+str(type(other)))

        res_terms_dict = {}

        for pauli,coeff in self.terms_dict.items():
            res_terms_dict[pauli] = res_terms_dict.get(pauli,0)-coeff
            if abs(res_terms_dict[pauli])<threshold:
                del res_terms_dict[pauli]
    
        for pauli,coeff in other.terms_dict.items():
            res_terms_dict[pauli] = res_terms_dict.get(pauli,0)+coeff
            if abs(res_terms_dict[pauli])<threshold:
                del res_terms_dict[pauli]
        
        result = BoundQubitOperator(res_terms_dict)
        return result
    
    def __mul__(self,other):
        """
        Returns the product of the operator self and other.

        Parameters
        ----------
        other : int, float, complex or BoundQubitOperator
            A scalar or a BoundQubitOperator to multiply with the operator self.

        Returns
        -------
        result : BoundQubitOperator
            The product of the operator self and other.

        """

        if isinstance(other,(int,float,complex)):
            other = BoundQubitOperator({BoundQubitTerm():other})
        if not isinstance(other,BoundQubitOperator):
            raise TypeError("Cannot multipliy BoundQubitOperator and "+str(type(other)))

        res_terms_dict = {}

        for pauli1, coeff1 in self.terms_dict.items():
            for pauli2, coeff2 in other.terms_dict.items():
                curr_pauli, curr_coeff = pauli1*pauli2
                res_terms_dict[curr_pauli] = res_terms_dict.get(curr_pauli,0) + curr_coeff*coeff1*coeff2

        result = BoundQubitOperator(res_terms_dict)
        return result

    __radd__ = __add__
    __rmul__ = __mul__

    #
    # Inplace arithmetic
    #

    def __iadd__(self,other):
        """
        Adds other to the operator self.

        Parameters
        ----------
        other : int, float, complex or BoundQubitOperator
            A scalar or a BoundQubitOperator to add to the operator self.

        """

        if isinstance(other,(int,float,complex)):
            self.terms_dict[BoundQubitTerm()] = self.terms_dict.get(BoundQubitTerm(),0)+other
            return self
        if not isinstance(other,BoundQubitOperator):
            raise TypeError("Cannot add BoundQubitOperator and "+str(type(other)))

        for pauli,coeff in other.terms_dict.items():
            self.terms_dict[pauli] = self.terms_dict.get(pauli,0)+coeff
            if abs(self.terms_dict[pauli])<threshold:
                del self.terms_dict[pauli]       
        return self         

    def __isub__(self,other):
        """
        Substracts other from the operator self.

        Parameters
        ----------
        other : int, float, complex or BoundQubitOperator
            A scalar or a BoundQubitOperator to substract from the operator self.

        """

        if isinstance(other,(int,float,complex)):
            self.terms_dict[BoundQubitTerm()] = self.terms_dict.get(BoundQubitTerm(),0)-other
            return self
        if not isinstance(other,BoundQubitOperator):
            raise TypeError("Cannot add BoundQubitOperator and "+str(type(other)))

        for pauli,coeff in other.terms_dict.items():
            self.terms_dict[pauli] = self.terms_dict.get(pauli,0)-coeff
            if abs(self.terms_dict[pauli])<threshold:
                del self.terms_dict[pauli]  
        return self
    
    def __imul__(self,other):
        """
        Multiplys other to the operator self.

        Parameters
        ----------
        other : int, float, complex or BoundQubitOperator
            A scalar or a BoundQubitOperator to multiply with the operator self.

        """

        if isinstance(other,(int,float,complex)):
            #other = BoundQubitOperator({BoundQubitTerm():other})
            for term in self.terms_dict:
                self.terms_dict[term] *= other
            return self

        if not isinstance(other,BoundQubitOperator):
            raise TypeError("Cannot multipliy BoundQubitOperator and "+str(type(other)))

        res_terms_dict = {}

        for pauli1, coeff1 in self.terms_dict.items():
            for pauli2, coeff2 in other.terms_dict.items():
                curr_pauli, curr_coeff = pauli1*pauli2
                res_terms_dict[curr_pauli] = res_terms_dict.get(curr_pauli,0) + curr_coeff*coeff1*coeff2

        self.terms_dict = res_terms_dict  
        return self  

    #
    # Substitution
    #

    def subs(self, subs_dict):
        """
        
        Parameters
        ----------
        subs_dict : dict
            A dictionary with indices (int) as keys and numbers (int, float, complex) as values.

        Returns
        -------
        result : BoundQubitOperator
            The resulting BoundQubitOperator.
        
        """

        res_terms_dict = {}

        for pauli, coeff in self.terms_dict.items():
            curr_pauli, curr_coeff = pauli.subs(subs_dict)
            res_terms_dict[curr_pauli] = res_terms_dict.get(curr_pauli,0) + curr_coeff*coeff

        result = BoundQubitOperator(res_terms_dict)
        return result
    
    #
    # Miscellaneous
    #

    def apply_threshold(self,threshold):
        """
        Removes all Pauli terms with coefficient absolute value below the specified threshold.

        Parameters
        ----------
        threshold : float
            The threshold for the coefficients of the Pauli terms.

        """

        delete_list = []
        for pauli,coeff in self.terms_dict.items():
            if abs(coeff)<threshold:
                delete_list.append(pauli)
        for pauli in delete_list:
            del self.terms_dict[pauli]

    def to_sparse_matrix(self):
        """
        Returns a matrix representing the operator.
    
        Returns
        -------
        M : scipy.sparse.csr_matrix
            A sparse matrix representing the operator.

        """

        import scipy.sparse as sp
        from scipy.sparse import kron as TP, csr_matrix

        I = csr_matrix([[1,0],[0,1]])

        def get_matrix(P):
            if P=="I":
                return csr_matrix([[1,0],[0,1]])
            if P=="X":
                return csr_matrix([[0,1],[1,0]])
            if P=="Y":
                return csr_matrix([[0,-1j],[1j,0]])
            else:
                return csr_matrix([[1,0],[0,-1]])

        def recursive_TP(keys,factor_dict):
            if len(keys)==1:
                return get_matrix(factor_dict.get(keys[0],"I"))
            return TP(get_matrix(factor_dict.get(keys.pop(0),"I")),recursive_TP(keys,factor_dict))

        factor_dicts = []
        coeffs = []

        keys = set()
        for pauli,coeff in self.terms_dict.items():
            curr_dict = pauli.factor_dict
            keys.update(set(curr_dict.keys()))
            factor_dicts.append(curr_dict)    
            coeffs.append(coeff)

        keys = set()
        for item in factor_dicts:
            keys.update(set(item.keys()))
        keys = sorted(keys)
        dim = len(keys)

        m = len(coeffs)
        M = sp.csr_matrix((2**dim, 2**dim))
        for k in range(m):
            M += complex(coeffs[k])*recursive_TP(keys.copy(),factor_dicts[k])

        return M

    def ground_state_energy(self):
        """
        Calculates the ground state energy (i.e., the minimum eigenvalue) of the operator classically.
    
        Returns
        -------
        E : float
            The ground state energy. 

        """

        from scipy.sparse.linalg import eigsh

        M = self.to_sparse_matrix()
        # Compute the smallest eigenvalue
        eigenvalues, _ = eigsh(M, k=1, which='SA')  # 'SA' stands for smallest algebraic
        E = eigenvalues[0]

        return E
    
    #
    # Partitions 
    #

    # Commutativity: Partitions the BoundQubitOperator into BoundQubitOperators with pairwise commuting BoundQubitTerms
    def commuting_groups(self):
        r"""
        Partitions the QubitOperator into QubitOperators with pairwise commuting terms. That is,

        .. math::

            H = \sum_{i=1}^mH_i

        where the terms in each $H_i$ are mutually commuting.

        Returns
        -------
        groups : list[QubitOperator]
            The partition of the Hamiltonian.
        
        """

        groups = [] # Groups of commuting BoundQubitTerms 

        # Sorted insertion heuristic https://quantum-journal.org/papers/q-2021-01-20-385/pdf/
        sorted_terms = sorted(self.terms_dict.items(), key=lambda item: abs(item[1]), reverse=True)

        for pauli,coeff in sorted_terms:

            commute_bool = False
            if len(groups) > 0:
                for group in groups:
                    for pauli_,coeff_ in group.terms_dict.items():
                        commute_bool = pauli_.commute(pauli)
                        if not commute_bool:
                            break
                    if commute_bool:
                        group.terms_dict[pauli]=coeff
                        break
            if len(groups)==0 or not commute_bool: 
                groups.append(BoundQubitOperator({pauli:coeff}))

        return groups

    # Qubit-wise commutativity: Partitions the BoundQubitOperator into BoundQubitOperators with pairwise qubit-wise commuting BoundQubitTerms
    def commuting_qw_groups(self, show_bases=False):
        r"""
        Partitions the BoundQubitOperator into BoundQubitOperators with pairwise qubit-wise commuting terms. That is,

        .. math::

            H = \sum_{i=1}^mH_i

        where the terms in each $H_i$ are mutually qubit-wise commuting.

        Returns
        -------
        groups : list[BoundQubitOperator]
            The partition of the Hamiltonian.
        
        """

        groups = [] # Groups of qubit-wise commuting BoundQubitTerms
        bases = [] # Bases as BoundQubitTerms

        # Sorted insertion heuristic https://quantum-journal.org/papers/q-2021-01-20-385/pdf/
        sorted_terms = sorted(self.terms_dict.items(), key=lambda item: abs(item[1]), reverse=True)

        for pauli,coeff in sorted_terms:

            commute_bool = False
            if len(groups)>0:
                n = len(groups)
                for i in range(n):
                    commute_bool = bases[i].commute_qw(pauli)
                    if commute_bool:
                        bases[i].update(pauli.factor_dict)
                        groups[i].terms_dict[pauli]=coeff
                        break
            if len(groups)==0 or not commute_bool:
                groups.append(BoundQubitOperator({pauli:coeff}))
                bases.append(pauli.copy())

        if show_bases:
            return groups, bases
        else:
            return groups
    
    #
    # Measurement settings
    #

    def get_measurement(
        self,
        qarg,
        precision=0.01,
        backend=None,
        shots=1000000,
        compile=True,
        compilation_kwargs={},
        subs_dic={},
        precompiled_qc=None,
        measurement_data=None # measurement settings
    ):
        r"""
        This method returns the expected value of a Hamiltonian for the state of a quantum argument.

        Parameters
        ----------
        qarg : QuantumVariable, QuantumArray or list[QuantumVariable]
            The quantum argument to evaluate the Hamiltonian on.
        precision: float, optional
            The precision with which the expectation of the Hamiltonian is to be evaluated.
            The default is 0.01. The number of shots scales quadratically with the inverse precision.
        backend : BackendClient, optional
            The backend on which to evaluate the quantum circuit. The default can be
            specified in the file default_backend.py.
        shots : integer, optional
            The maximum amount of shots to evaluate the expectation of the Hamiltonian. 
            The default is 1000000.
        compile : bool, optional
            Boolean indicating if the .compile method of the underlying QuantumSession
            should be called before. The default is True.
        compilation_kwargs  : dict, optional
            Keyword arguments for the compile method. For more details check
            :meth:`QuantumSession.compile <qrisp.QuantumSession.compile>`. The default
            is ``{}``.
        subs_dic : dict, optional
            A dictionary of Sympy symbols and floats to specify parameters in the case
            of a circuit with unspecified, :ref:`abstract parameters<QuantumCircuit>`.
            The default is {}.
        circuit_preprocessor : Python function, optional
            A function which recieves a QuantumCircuit and returns one, which is applied
            after compilation and parameter substitution. The default is None.
        precompiled_qc : QuantumCircuit, optional
            A precompiled quantum circuit.
            
        Raises
        ------
        Exception
            If the containing QuantumSession is in a quantum environment, it is not
            possible to execute measurements.

        Returns
        -------
        float
            The expected value of the Hamiltonian.

        Examples
        --------

        We define a Hamiltonian, and measure its expected value for the state of a :ref:`QuantumVariable`.

        ::

            from qrisp import QuantumVariable, h
            from qrisp.operators import X,Y,Z
            qv = QuantumVariable(2)
            h(qv)
            H = Z(0)*Z(1)
            res = H.get_measurement(qv)
            print(res)
            #Yields 0.0

        We define a Hamiltonian, and measure its expected value for the state of a :ref:`QuantumArray`.

        ::

            from qrisp import QuantumVariable, QuantumArray, h
            from qrisp.operators import X,Y,Z
            qtype = QuantumVariable(2)
            q_array = QuantumArray(qtype, shape=(2))
            h(q_array)
            H = Z(0)*Z(1) + X(2)*X(3)
            res = H.get_measurement(q_array)
            print(res)
            #Yields 1.0

        """
        
        unbound_operator, qarg = self.unbind()
        
        return unbound_operator.get_measurement(qarg,
                                                precision=precision, 
                                                backend=backend, 
                                                shots=shots, 
                                                compile=compile, 
                                                compilation_kwargs=compilation_kwargs, 
                                                subs_dic=subs_dic,
                                                precompiled_qc=precompiled_qc, 
                                                measurement_data=measurement_data)

    def unbind(self):
        from qrisp.operators import QubitTerm, QubitOperator
        
        participating_qubits = []
        for term, coeff in self.terms_dict.items():
            for qb in term.factor_dict.keys():
                if qb not in participating_qubits:
                    participating_qubits.append(qb)
        
        index_inv = {participating_qubits[i] : i for i in range(len(participating_qubits))}
        
        unbound_operator = 0
        
        for term, coeff in self.terms_dict.items():
            new_term_factor_dict = {}
            for qb, factor in term.factor_dict.items():
                new_term_factor_dict[index_inv[qb]] = factor
            
            unbound_operator += QubitOperator({QubitTerm(new_term_factor_dict): coeff})
        
        return unbound_operator, participating_qubits
        
    
    #
    # Trotterization
    #
    

    def trotterization(self):
        r"""
        Returns a function for performing Hamiltonian simulation, i.e., approximately implementing the unitary operator $e^{itH}$ via Trotterization.

        Returns
        -------
        U : function 
            A Python function that implements the first order Suzuki-Trotter formula.
            Given a Hamiltonian $H=H_1+\dotsb +H_m$ the unitary evolution $e^{itH}$ is 
            approximated by 
            
            .. math::

                e^{itH}\approx U_1(t,N)=\left(e^{iH_1t/N}\dotsb e^{iH_mt/N}\right)^N

            This function receives the following arguments:

            * qarg : QuantumVariable 
                The quantum argument.
            * t : float, optional
                The evolution time $t$. The default is 1.
            * steps : int, optional
                The number of Trotter steps $N$. The default is 1.
            * iter : int, optional 
                The number of iterations the unitary $U_1(t,N)$ is applied. The default is 1.
        
        """

        def change_of_basis(factor_dict):
            for qubit, axis in factor_dict.items():
                if axis=="X":
                    h(qubit)
                if axis=="Y":
                    sx(qubit)

        groups, bases = self.commuting_qw_groups(show_bases=True)

        def trotter_step(qarg, t, steps):

            # qubit to apply gphase for identity term
            if isinstance(qarg,list):
                qubit = qarg[0][0]
            else:
                qubit = qarg[0]

            for index,basis in enumerate(bases):
                with conjugate(change_of_basis)(basis.factor_dict):
                    for term,coeff in groups[index].terms_dict.items():
                        term.simulate(coeff*t/steps, qubit)

        def U(qarg, t=1, steps=1, iter=1):
            merge([qarg])
            with IterationEnvironment(qarg.qs, iter*steps):
                trotter_step(qarg, t, steps)

        return U


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.core.gate_application_functions import h
from qrisp.alg_primitives.qft import QFT
from qrisp.core import QuantumArray
from qrisp.qtypes import QuantumBool
from qrisp.environments import conjugate, control
from qrisp.alg_primitives import demux

def qswitch(operand, case, case_function_list, method = "sequential"):
    """
    Executes a switch - case statement distinguishing between a list of
    given in-place functions.
        

    Parameters
    ----------
    operand : QuantumVariable
        The quantum argument on which to execute the case function.
    case : QuantumFloat
        A QuantumFloat specifying which case function should be executed.
    case_function_list : list[callable]
        A list of functions, performing some in-place operation on ``operand``.
    method : str, optional
        The compilation method. Available are ``parallel`` and ``sequential``. 
        ``parallel`` is exponentially fast but requires more temporary qubits.
        The default is "sequential".

    Examples
    --------
    
    We create some sample functions:

    >>> from qrisp import *
    >>> def f0(x): x += 1
    >>> def f1(x): inpl_mult(x, 3, treat_overflow = False)
    >>> def f2(x): pass
    >>> def f3(x): h(x[1])
    >>> case_function_list = [f0, f1, f2, f3]

    Create operand and case variable
    
    >>> operand = QuantumFloat(4)
    >>> operand[:] = 1
    >>> case = QuantumFloat(2)
    >>> h(case)

    Execute switch_case function
    
    >>> qswitch(operand, case, case_function_list)
    
    Simulate
    
    >>> print(multi_measurement([case, operand]))
    {(0, 2): 0.25, (1, 3): 0.25, (2, 1): 0.25, (3, 1): 0.125, (3, 3): 0.125}

    """
    
    
    if method == "sequential":
    
        for i in range(len(case_function_list)):
            with i == case:
                case_function_list[i](operand)
        
    elif method == "parallel":
        
        # Idea: Use demux function to move operand and enabling bool into QuantumArray
        # to execute cases in parallel.
        case_amount = len(case_function_list)
    
        # This QuantumArray acts as an addressable QRAM via the demux function
        enable = QuantumArray(qtype = QuantumBool(), shape = (case_amount,))
        enable[0].flip()
        
        qa = QuantumArray(qtype = operand, shape = ((case_amount,)))
    
        with conjugate(demux)(operand, case, qa, parallelize_qc = True):
            with conjugate(demux)(enable[0], case, enable, parallelize_qc = True):
                for i in range(case_amount):
                    with control(enable[i]):
                        case_function_list[i](qa[i])
                        
        qa.delete()
        
        enable[0].flip()
        enable.delete()
    
    else:
        raise Exception("Don't know compile method {method} for switch-case structure.")


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import numpy as np

from qrisp.qtypes.quantum_float import QuantumFloat
from qrisp.alg_primitives.arithmetic import multi_controlled_U_g, hybrid_mult
from qrisp.alg_primitives import QFT, h, cx, swap, mcx
from qrisp.environments import conjugate, control
from qrisp.alg_primitives.arithmetic.modular_arithmetic.mod_tools import modinv, montgomery_encoder

def QREDC(t, N, m):
    
    u = QuantumFloat(0, qs = t.qs, name = "u*")
    
    for i in range(m):
        
        h(t[0])
        
        transfer_lsb(t, u)
        
        t.exp_shift(-1)
        multi_controlled_U_g(t, [u[-1]], -N/2)

    
    QFT(t, inv = True, exec_swap = False)
    
    with conjugate(QFT)(t[:-1], exec_swap = False):
        
        multi_controlled_U_g(t[:-1], [t.sign()], N)
    
    cx(t[0], t.sign())
    
    sgn = t.reg.pop(-1)
    t.size -= 1
    t.signed = False
    
    u.reg.insert(len(u), sgn)
    u.size += 1
    u.mshape[1] += 1
    
    
    t.m = t.m - m
    
    return t, u


def find_optimal_m(b, N):
    m = 0
    n = int(np.ceil(np.log2(N)))
    
    while True:
        b_trial = montgomery_encoder(b, 2**m, N)%N
        if (N-1)*b_trial < N*2**(m):
            return m
        m += 1

def montgomery_mod_semi_mul(a, b, output_qg = None, permeable_if_zero = False):
    
    
    N = a.modulus
    
    b = b%N
    
    m = find_optimal_m(b, N)
    
    if b == 0:
        if output_qg is None:
            return a.duplicate()
        else:
            return output_qg
    if b == 1:
        if output_qg is None:
            return a.duplicate(init = True)
        else:
            output_qg += a
            return output_qg
    
    m = find_optimal_m(b, N)
    b = montgomery_encoder(b, 2**m, N)%N
    
    
    if output_qg is None:
        t = QuantumFloat(a.size+m, signed = True)
        
        h(t)
    else:
        if output_qg.modulus != N:
            raise Exception("Output QuantumModulus has incompatible modulus")
        
        output_qg.extend(m, 0)
        output_qg.add_sign()
        
        QFT(output_qg, exec_swap = False)
        
        t = output_qg
    
    
    for i in range(a.size):
        multi_controlled_U_g(t, [a[i]], (2**i*b))
    
    from qrisp import QuantumModulus
    t.__class__ = QuantumModulus
    t.modulus = a.modulus
    t.m = a.m + m
    t.inpl_adder = a.inpl_adder
    
    
    
    return montgomery_red(t, a, b, N, m, permeable_if_zero = permeable_if_zero)


def montgomery_red(t, a, b, N, m, permeable_if_zero = False):
    
    t, u = QREDC(t, N, m)
    
    with conjugate(QFT)(u, exec_swap = False):
        if isinstance(b, QuantumFloat):
            
            hybrid_mult(a, 
                        b, 
                        output_qf=u, 
                        init_op = None, 
                        terminal_op = None, 
                        cl_factor = -modinv(N, 2**(m+1)))
            
        else:
            for i in range(len(a)):
                multi_controlled_U_g(u, [a[i]], -((2**i*b))*modinv(N, 2**(m+1)))

    if permeable_if_zero:    
        mcx(list(a) + [t[0]],  
            u[-1], ctrl_state = "0"*len(a) + "1",
            method = "balauca")
    
    u.delete(verify = False)
    
    return t

from qrisp import merge
def montgomery_mod_mul(a, b, output_qg = None):
    
    m = int(np.ceil(np.log2((a.modulus-1)**2)+1)) - a.size
    
    if a.modulus != b.modulus:
        raise Exception("Tried to multiply two QuantumModulus with differing modulus")
        
    if output_qg is None:
        t = QuantumFloat(a.size + m, signed = True)
        h(t)
        
    else:
        if output_qg.modulus != a.modulus:
            raise Exception("Output QuantumModulus has incompatible modulus")
        
        merge(output_qg.qs, a.qs)
        output_qg.extend(m, 0)
        output_qg.add_sign()
        output_qg.reg.insert(0, output_qg.reg.pop(-1))
        
        QFT(output_qg, exec_swap = False)
        
        t = output_qg
    
    t = hybrid_mult(a, b, init_op = None, terminal_op = None, output_qf = t)
    
    from qrisp import QuantumModulus
    t.__class__ = QuantumModulus
    t.modulus = a.modulus
    t.m = (a.m + b.m)
    t.inpl_adder = a.inpl_adder
    
    t = montgomery_red(t, a, b, a.modulus, m)
    
    return t

from qrisp import invert
from qrisp.environments import custom_control

@custom_control
def qft_semi_cl_inpl_mult(a, X, ctrl = None, treat_invalid = False):
    
    X = X%a.modulus
    
    if X == 0:
        raise Exception("Tried to perform in-place multiplication with 0 (not invertible)")
    if X == 1:
        return a
    
    tmp = a.duplicate(qs = a.qs)
    
    from qrisp import multi_measurement, less_than, redirect_qfunction, QuantumModulus
    
    if treat_invalid:
        a.__class__ = QuantumFloat
        reduced = a < a.modulus
        a.__class__ = QuantumModulus
        ctrl = [ctrl, reduced[0]]
    
    if ctrl is not None:
        with control(ctrl, invert = True):
            swap(tmp, a)
            
    tmp = montgomery_mod_semi_mul(a, 
                                  X, 
                                  output_qg = tmp,
                                  permeable_if_zero = ctrl is not None)
    
    if ctrl is not None:
        with control(ctrl, invert = True):
            swap(tmp, a)
    
    with invert():
        a = montgomery_mod_semi_mul(tmp, 
                                    modinv(X, a.modulus)%a.modulus, 
                                    output_qg = a, 
                                    permeable_if_zero = ctrl is not None)
        
    if ctrl is not None:
        with control(ctrl, invert = False):
            swap(tmp, a)
    else:
        swap(tmp, a)
    
    tmp.delete(verify = False)
    
    if treat_invalid:
        a.__class__ = QuantumFloat
        redirect_qfunction(less_than)(a,a.modulus,target= reduced)
        a.__class__ = QuantumModulus
        reduced.delete(verify = False)
    
    
    return a



            
    
    

def transfer_lsb(from_qv, to_qv):
    lsb = from_qv.reg.pop(0)
    from_qv.size -= 1
    from_qv.msize -= 1
    from_qv.mshape[0] += 1
    from_qv.exponent += 1

    to_qv.reg.insert(len(to_qv), lsb)
    to_qv.size += 1
    to_qv.msize += 1
    to_qv.mshape[1] += 1


"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""
from qrisp.operators.hamiltonian_tools import group_up_terms
from qrisp.operators.hamiltonian import Hamiltonian
from qrisp.operators.qubit.qubit_term import QubitTerm
from qrisp.operators.qubit.measurement import get_measurement
from qrisp.operators.qubit.commutativity_tools import gaussian_elimination_mod2, inverse_mod2, construct_change_of_basis
from qrisp import cx, cz, h, s, x, sx_dg, IterationEnvironment, conjugate, merge

import sympy as sp
import numpy as np

threshold = 1e-9

#
# QubitOperator
#

class QubitOperator(Hamiltonian):
    r"""
    This class provides an efficient implementation of QubitOperators, i.e.
    Operators, that act on a qubit space :math:`(\mathbb{C}^2)^{\otimes n}`.
    Supported are operators of the following form:
    
    .. math::
        
        O=\sum\limits_{j}\alpha_j O_j 
        
    where :math:`O_j=\bigotimes_i m_i^j` is a product of the following operators:
    
    .. list-table::
       :header-rows: 1
       :widths: 20 40 40
    
       * - Operator
         - Ket-Bra Realization
         - Description
       * - $X$
         - :math:`\ket{0}\bra{1} + \ket{1}\bra{0}`
         - Pauli-X operator (bit flip)
       * - $Y$
         - :math:`-i\ket{0}\bra{1} + i\ket{1}\bra{0}`
         - Pauli-Y operator (bit flip with phase)
       * - $Z$
         - :math:`\ket{0}\bra{0} - \ket{1}\bra{1}`
         - Pauli-Z operator (phase flip)
       * - $A$
         - :math:`\ket{0}\bra{1}`
         - Annihilation operator
       * - $C$
         - :math:`\ket{1}\bra{0}`
         - Creation operator
       * - $P_0$
         - :math:`\ket{0}\bra{0}`
         - Projector onto the :math:`\ket{0}` state
       * - $P_1$
         - :math:`\ket{1}\bra{1}`
         - Projector onto the :math:`\ket{1}` state
       * - $I$
         - :math:`\ket{1}\bra{1} + \ket{0}\bra{0}`
         - Identity operator
    
    If you already have some experience you might wonder why to include the
    non-Pauli operators - after all they can be represented as a linear
    combination of ``X``, ``Y`` and ``Z``. 
    
    .. math::
        
        \begin{align}
        A_0 C_1 &= (X_0 - i Y_0)(X_1 + Y_1)/4 \\
        & = (X_0X_1 + X_0Y_1 - Y_0X_1 + Y_0Y_1)/4
        \end{align}
    
    Recently a much more efficient method of simulating ``A`` and ``C`` `has 
    been proposed by Kornell and Selinger <https://arxiv.org/abs/2310.12256>`_,
    which avoids decomposing these Operators into Paulis strings
    but instead simulates 
    
    .. math::
        H = A_0C_1 + h.c. 
        
    within a single step.
    This idea is deeply integrated into the Operators module of Qrisp. For an
    example circuit see below.

    Examples
    --------
    
    A QubitOperator can be specified conveniently in terms of arithmetic 
    combinations of the mentioned operators:

    ::
        
        from qrisp.operators.qubit import X,Y,Z,A,C,P0,P1

        H = 1+2*X(0)+3*X(0)*Y(1)*A(2)+C(4)*P1(0)
        H

    Yields $1 + P^1_0C_4 + 2X_0 + 3X_0Y_1A_2$.

    We create a QubitOperator and perform Hamiltonian simulation via :meth:`trotterization <QubitOperator.trotterization>`:

    ::
        
        from sympy import Symbol
        from qrisp.operators import A,C,Z,Y
        from qrisp import QuantumVariable
        O = A(0)*C(1)*Z(2)*A(3) + Y(3)
        U = O.trotterization()
        qv = QuantumVariable(4)
        t = Symbol("t")
        U(qv, t = t)
        
    >>> print(qv.qs)
    QuantumCircuit:
    ---------------
              ┌───┐                                                                »
        qv.0: ┤ X ├────────────o──────────────────────────────────────o────────────»
              └─┬─┘┌───┐       │                                      │       ┌───┐»
        qv.1: ──┼──┤ X ├───────■──────────────────────────────────────■───────┤ X ├»
                │  └─┬─┘       │                                      │       └─┬─┘»
        qv.2: ──┼────┼─────────┼────■────────────────────────────■────┼─────────┼──»
                │    │  ┌───┐  │  ┌─┴─┐     ┌────────────┐     ┌─┴─┐  │  ┌───┐  │  »
        qv.3: ──■────■──┤ H ├──┼──┤ X ├──■──┤ Rz(-0.5*t) ├──■──┤ X ├──┼──┤ H ├──■──»
                        └───┘┌─┴─┐└───┘┌─┴─┐├───────────┬┘┌─┴─┐└───┘┌─┴─┐└───┘     »
    hs_anc.0: ───────────────┤ X ├─────┤ X ├┤ Rz(0.5*t) ├─┤ X ├─────┤ X ├──────────»
                             └───┘     └───┘└───────────┘ └───┘     └───┘          »
    «          ┌───┐                            
    «    qv.0: ┤ X ├────────────────────────────
    «          └─┬─┘                            
    «    qv.1: ──┼──────────────────────────────
    «            │                              
    «    qv.2: ──┼──────────────────────────────
    «            │  ┌────┐┌────────────┐┌──────┐
    «    qv.3: ──■──┤ √X ├┤ Rz(-2.0*t) ├┤ √Xdg ├
    «               └────┘└────────────┘└──────┘
    «hs_anc.0: ─────────────────────────────────
    «                                           
    Live QuantumVariables:
    ----------------------
    QuantumVariable qv
    
    Call the simulator:
        
    >>> print(qv.get_measurement(subs_dic = {t : 0.5}))
    {'0000': 0.77015, '0001': 0.22985}

    """

    def __init__(self, terms_dict={}):
        self.terms_dict = dict(terms_dict)

    def len(self):
        return len(self.terms_dict)
    
    #
    # Printing
    #

    def _repr_latex_(self):
        # Convert the sympy expression to LaTeX and return it
        expr = self.to_expr()
        return f"${sp.latex(expr)}$"
    
    def __str__(self):
        # Convert the sympy expression to a string and return it
        expr = self.to_expr()
        return str(expr)

    def __repr__(self):
        # Convert the sympy expression to a string and return it
        return str(self)
    
    def to_expr(self):
        """
        Returns a SymPy expression representing the operator.

        Returns
        -------
        expr : sympy.expr
            A SymPy expression representing the operator.

        """
        
        expr = 0  
        for term, coeff in self.terms_dict.items():
            expr += coeff*term.to_expr()
        return expr

    #
    # Arithmetic
    #

    def __pow__(self, e):
        res = 1
        for i in range(e):
            res = res * self
        return res    

    def __add__(self,other):
        """
        Returns the sum of the operator self and other.

        Parameters
        ----------
        other : int, float, complex or QubitOperator
            A scalar or a QubitOperator to add to the operator self.

        Returns
        -------
        result : QubitOperator
            The sum of the operator self and other.

        """

        if isinstance(other,(int,float,complex)):
            other = QubitOperator({QubitTerm():other})
        if not isinstance(other,QubitOperator):
            raise TypeError("Cannot add QubitOperator and "+str(type(other)))

        res_terms_dict = {}

        for term,coeff in self.terms_dict.items():
            res_terms_dict[term] = res_terms_dict.get(term,0)+coeff
            if abs(res_terms_dict[term])<threshold:
                del res_terms_dict[term]
    
        for term,coeff in other.terms_dict.items():
            res_terms_dict[term] = res_terms_dict.get(term,0)+coeff
            if abs(res_terms_dict[term])<threshold:
                del res_terms_dict[term]
        
        result = QubitOperator(res_terms_dict)
        return result
    
    def __sub__(self,other):
        """
        Returns the difference of the operator self and other.

        Parameters
        ----------
        other : int, float, complex or QubitOperator
            A scalar or a QubitOperator to substract from the operator self.

        Returns
        -------
        result : QubitOperator
            The difference of the operator self and other.

        """

        if isinstance(other,(int,float,complex)):
            other = QubitOperator({QubitTerm():other})
        if not isinstance(other,QubitOperator):
            raise TypeError("Cannot substract QubitOperator and "+str(type(other)))

        res_terms_dict = {}

        for term, coeff in self.terms_dict.items():
            res_terms_dict[term] = res_terms_dict.get(term,0)+coeff
            if abs(res_terms_dict[term])<threshold:
                del res_terms_dict[term]
    
        for term,coeff in other.terms_dict.items():
            res_terms_dict[term] = res_terms_dict.get(term,0)-coeff
            if abs(res_terms_dict[term])<threshold:
                del res_terms_dict[term]
        
        result = QubitOperator(res_terms_dict)
        return result
    
    def __rsub__(self,other):
        """
        Returns the difference of the operator other and self.

        Parameters
        ----------
        other : int, float, complex or QubitOperator
            A scalar or a QubitOperator to substract the operator self from.

        Returns
        -------
        result : QubitOperator
            The difference of the operator other and self.

        """

        if isinstance(other,(int,float,complex)):
            other = QubitOperator({QubitTerm():other})
        if not isinstance(other,QubitOperator):
            raise TypeError("Cannot substract QubitOperator and "+str(type(other)))

        res_terms_dict = {}

        for term,coeff in self.terms_dict.items():
            res_terms_dict[term] = res_terms_dict.get(term,0)-coeff
            if abs(res_terms_dict[term])<threshold:
                del res_terms_dict[term]
    
        for term,coeff in other.terms_dict.items():
            res_terms_dict[term] = res_terms_dict.get(term,0)+coeff
            if abs(res_terms_dict[term])<threshold:
                del res_terms_dict[term]
        
        result = QubitOperator(res_terms_dict)
        return result

    def __mul__(self,other):
        """
        Returns the product of the operator self and other.

        Parameters
        ----------
        other : int, float, complex or QubitOperator
            A scalar or a QubitOperator to multiply with the operator self.

        Returns
        -------
        result : QubitOperator
            The product of the operator self and other.

        """

        if isinstance(other,(int,float,complex)):
            other = QubitOperator({QubitTerm():other})
        if not isinstance(other,QubitOperator):
            raise TypeError("Cannot multipliy QubitOperator and "+str(type(other)))

        res_terms_dict = {}

        for term1, coeff1 in self.terms_dict.items():
            for term2, coeff2 in other.terms_dict.items():
                curr_term, curr_coeff = term1*term2
                res_terms_dict[curr_term] = res_terms_dict.get(curr_term,0) + curr_coeff*coeff1*coeff2

        result = QubitOperator(res_terms_dict)
        return result

    __radd__ = __add__
    __rmul__ = __mul__

    #
    # Inplace arithmetic
    #

    def __iadd__(self,other):
        """
        Adds other to the operator self.

        Parameters
        ----------
        other : int, float, complex or QubitOperator
            A scalar or a QubitOperator to add to the operator self.

        """

        if isinstance(other,(int,float,complex)):
            self.terms_dict[QubitTerm()] = self.terms_dict.get(QubitTerm(),0)+other
            return self
        if not isinstance(other,QubitOperator):
            raise TypeError("Cannot add QubitOperator and "+str(type(other)))

        for term,coeff in other.terms_dict.items():
            self.terms_dict[term] = self.terms_dict.get(term,0)+coeff
            if abs(self.terms_dict[term])<threshold:
                del self.terms_dict[term]       
        return self         

    def __isub__(self,other):
        """
        Substracts other from the operator self.

        Parameters
        ----------
        other : int, float, complex or QubitOperator
            A scalar or a QubitOperator to substract from the operator self.

        """

        if isinstance(other,(int,float,complex)):
            self.terms_dict[QubitTerm()] = self.terms_dict.get(QubitTerm(),0)-other
            return self
        if not isinstance(other,QubitOperator):
            raise TypeError("Cannot add QubitOperator and "+str(type(other)))

        for term,coeff in other.terms_dict.items():
            self.terms_dict[term] = self.terms_dict.get(term,0)-coeff
            if abs(self.terms_dict[term])<threshold:
                del self.terms_dict[term]  
        return self
    
    def __imul__(self,other):
        """
        Multiplys other to the operator self.

        Parameters
        ----------
        other : int, float, complex or QubitOperator
            A scalar or a QubitOperator to multiply with the operator self.

        """

        if isinstance(other,(int,float,complex)):
            #other = QubitOperator({QubitTerm():other})
            for term in self.terms_dict:
                self.terms_dict[term] *= other
            return self

        if not isinstance(other,QubitOperator):
            raise TypeError("Cannot multipliy QubitOperator and "+str(type(other)))

        res_terms_dict = {}

        for term1, coeff1 in self.terms_dict.items():
            for term2, coeff2 in other.terms_dict.items():
                curr_term, curr_coeff = term1*term2
                res_terms_dict[curr_term] = res_terms_dict.get(curr_term,0) + curr_coeff*coeff1*coeff2

        self.terms_dict = res_terms_dict    
        return self

    #
    # Substitution
    #

    def subs(self, subs_dict):
        """
        
        Parameters
        ----------
        subs_dict : dict
            A dictionary with indices (int) as keys and numbers (int, float, complex) as values.

        Returns
        -------
        result : QubitOperator
            The resulting QubitOperator.
        
        """

        res_terms_dict = {}

        for term, coeff in self.terms_dict.items():
            curr_term, curr_coeff = term.subs(subs_dict)
            res_terms_dict[curr_term] = res_terms_dict.get(curr_term,0) + curr_coeff*coeff

        return QubitOperator(res_terms_dict)

    #
    # Miscellaneous
    #
    
    def find_minimal_qubit_amount(self):
        indices = sum([list(term.factor_dict.keys()) for term in self.terms_dict.keys()], [])
        if len(indices) == 0:
            return 0
        return max(indices)+1
    
    def commutator(self, other):
        """
        Computes the commutator.
        
        .. math::
            
            [A,B] = AB - BA

        Parameters
        ----------
        other : QubitOperator
            The second argument of the commutator.

        Returns
        -------
        commutator : QubitOperator
            The commutator operator.
            
        Examples
        --------
        
        We compute the commutator of a ladder operator with a Pauli string.
        
        >>> from qrisp.operators import A,C,X,Z
        >>> O_0 = A(0)*C(1)*A(2)
        >>> O_1 = Z(0)*X(1)*X(1)
        >>> print(O_0.commutator(O_1))
        2*A_0*C_1*A_2
        

        """
        
        res = 0
        
        for term_self, coeff_self in self.terms_dict.items():
            for term_other, coeff_other in other.terms_dict.items():
                res += coeff_self*coeff_other*term_self.commutator(term_other)
                
        min_coeff_self = min([abs(coeff) for coeff in self.terms_dict.values()])
        min_coeff_other = min([abs(coeff) for coeff in other.terms_dict.values()])
        
        res.apply_threshold(min_coeff_self*min_coeff_other/2)
        
        return res
        
    

    def apply_threshold(self,threshold):
        """
        Removes all terms with coefficient absolute value below the specified threshold.

        Parameters
        ----------
        threshold : float
            The threshold for the coefficients of the terms.

        """

        delete_list = []
        for term,coeff in self.terms_dict.items():
            if abs(coeff)<threshold:
                delete_list.append(term)
        for term in delete_list:
            del self.terms_dict[term]

    def to_sparse_matrix(self, factor_amount = None):
        """
        Returns a matrix representing the operator.
    
        Returns
        -------
        M : scipy.sparse.csr_matrix
            A sparse matrix representing the operator.

        """

        import scipy.sparse as sp
        from scipy.sparse import kron as TP, csr_matrix

        def get_matrix(P):
            if P=="I":
                return csr_matrix([[1,0],[0,1]])
            if P=="X":
                return csr_matrix([[0,1],[1,0]])
            if P=="Y":
                return csr_matrix([[0,-1j],[1j,0]])
            if P == "Z":
                return csr_matrix([[1,0],[0,-1]])
            if P == "A":
                return csr_matrix([[0,1],[0,0]])
            if P == "C":
                return csr_matrix([[0,0],[1,0]])
            if P == "P0":
                return csr_matrix([[1,0],[0,0]])
            if P == "P1":
                return csr_matrix([[0,0],[0,1]])

        def recursive_TP(keys,term_dict):
            if len(keys)==1:
                return get_matrix(term_dict.get(keys[0],"I"))
            return TP(get_matrix(term_dict.get(keys.pop(0),"I")),recursive_TP(keys,term_dict))

        term_dicts = []
        coeffs = []

        participating_indices = set()
        for term,coeff in self.terms_dict.items():
            curr_dict = term.factor_dict
            term_dicts.append(curr_dict)    
            coeffs.append(coeff)
            participating_indices = participating_indices.union(term.non_trivial_indices())

        if factor_amount is None:
            if len(participating_indices):
                factor_amount = max(participating_indices) + 1
            else:
                res = 1
                for coeff in coeffs:
                    res *= coeff
                M = sp.csr_matrix((1,1))
                M[0,0] = res
                return M
        elif participating_indices and factor_amount < max(participating_indices):
            raise Exception("Tried to compute Hermitian matrix with factor_amount variable lower than the largest factor index")

        keys = list(range(factor_amount))
        
        dim = len(keys)

        m = len(coeffs)
        M = sp.csr_matrix((2**dim, 2**dim))
        for k in range(m):
            M += complex(coeffs[k])*recursive_TP(keys.copy(),term_dicts[k])
        # res = ((M + M.transpose().conjugate())/2)
        # res.sum_duplicates()
        return M
    
    def to_array(self, factor_amount = None):
        """
        Returns a numpy array describing the operator.

        Parameters
        ----------
        factor_amount : int, optional
            The amount of factors to represent this array. The array will have 
            the dimension $2^n \times 2^n$, where n is the amount of factors. 
            By default the minimal number is chosen.

        Returns
        -------
        np.ndarray
            The array describing the operator.

        """
        return self.to_sparse_matrix(factor_amount).todense()
    
    def to_pauli(self):
        """
        Returns an equivalent operator, which however only contains Pauli factors.

        Returns
        -------
        QubitOperator
            An operator that contains only Pauli-Factor.

        Examples
        --------
        
        We create a QubitOperator containing A and C terms and convert it to a
        Pauli based representation.
        
        >>> from qrisp.operators import A,C,Z
        >>> H = A(0)*C(1)*Z(2)
        >>> print(H.to_pauli())
        0.25*X_0*X_1*Z_2 + 0.25*I*X_0*Y_1*Z_2 - 0.25*I*Y_0*X_1*Z_2 + 0.25*Y_0*Y_1*Z_2
            
        """
        
        res = 0
        for term, coeff in self.terms_dict.items():
            res += coeff*term.to_pauli()
        if isinstance(res, (float, int)):
            return QubitOperator({QubitTerm({}) : res})
        return res
    
    def adjoint(self):
        """
        Returns an the adjoint operator.

        Returns
        -------
        QubitOperator
            The adjoint operator.

        Examples
        --------
        
        We create a QubitOperator and inspect it' adjoint.
        
        >>> from qrisp.operators import A,C,Z
        >>> H = A(0)*C(1)*Z(2)
        >>> print(H.adjoint())
        C_0*A_1*Z_2
        """
        new_terms_dict = {}
        for term, coeff in self.terms_dict.items():
            new_terms_dict[term.adjoint()] = np.conjugate(coeff)
        return QubitOperator(new_terms_dict)
    
    def hermitize(self):
        """
        Returns the hermitian part of self.
        
        $H = (O + O^\dagger)/2$

        Returns
        -------
        QubitOperator
            The hermitian part.

        """
        return 0.5*(self + self.adjoint())
        
    def eliminate_ladder_conjugates(self):
        new_terms_dict = {}
        for term, coeff in self.terms_dict.items():
            for factor in term.factor_dict.values():
                if factor in ["A", "C"]:
                    break
            else:
                new_terms_dict[term] = coeff
                continue
            
            if term.adjoint() in new_terms_dict:
                new_terms_dict[term.adjoint()] += coeff
            else:
                new_terms_dict[term] = coeff
        
        return QubitOperator(new_terms_dict)
        
    
    def ground_state_energy(self):
        """
        Calculates the ground state energy (i.e., the minimum eigenvalue) of the operator classically.
    
        Returns
        -------
        E : float
            The ground state energy. 

        """

        from scipy.sparse.linalg import eigsh

        M = self.hermitize().to_sparse_matrix()
        # Compute the smallest eigenvalue
        eigenvalues, _ = eigsh(M, k=1, which='SA')  # 'SA' stands for smallest algebraic
        E = eigenvalues[0]

        return E
    
    #
    # Partitions 
    #

    # Commutativity: Partitions the QubitOperator into QubitOperators with pairwise commuting QubitTerms
    def commuting_groups(self):
        r"""
        Partitions the QubitOperator into QubitOperators with pairwise commuting terms. That is,

        .. math::

            H = \sum_{i=1}^mH_i

        where the terms in each $H_i$ are mutually commuting.

        Returns
        -------
        groups : list[QubitOperator]
            The partition of the Hamiltonian.
        
        """

        groups = [] # Groups of commuting QubitTerms 

        # Sorted insertion heuristic https://quantum-journal.org/papers/q-2021-01-20-385/pdf/
        sorted_terms = sorted(self.terms_dict.items(), key=lambda item: abs(item[1]), reverse=True)

        for term,coeff in sorted_terms:

            commute_bool = False
            if len(groups) > 0:
                for group in groups:
                    for term_,coeff_ in group.terms_dict.items():
                        commute_bool = term_.commute(term)
                        if not commute_bool:
                            break
                    if commute_bool:
                        group.terms_dict[term]=coeff
                        break
            if len(groups)==0 or not commute_bool: 
                groups.append(QubitOperator({term:coeff}))

        return groups
    
    def group_up(self, group_denominator):
        term_groups = group_up_terms(self, group_denominator)
        groups = []
        for term_group in term_groups:
            H = QubitOperator({term : self.terms_dict[term] for term in term_group})
            groups.append(H)
            
        return groups
        

    # Qubit-wise commutativity: Partitions the QubitOperator into QubitOperators with pairwise qubit-wise commuting QubitTerms
    def commuting_qw_groups(self, show_bases=False, use_graph_coloring = True):
        r"""
        Partitions the QubitOperator into QubitOperators with pairwise qubit-wise commuting terms. That is,

        .. math::

            H = \sum_{i=1}^mH_i

        where the terms in each $H_i$ are mutually qubit-wise commuting.

        Returns
        -------
        groups : list[QubitOperator]
            The partition of the Hamiltonian.
        
        """

        groups = [] # Groups of qubit-wise commuting QubitTerms
        bases = [] # Bases as termTerms

        if use_graph_coloring:        
            
            term_groups = group_up_terms(self, lambda a, b : a.commute_qw(b))
            for term_group in term_groups:
                H = QubitOperator({term : self.terms_dict[term] for term in term_group})
                groups.append(H)
                
                if show_bases:
                    factor_dict = {}
                    
                    for term in term_group:
                        for index, factor in term.factor_dict.items():
                            if factor in ["X", "Y", "Z"]:
                                factor_dict[index] = factor
                    
                    bases.append(QubitTerm(factor_dict))
                    
            if show_bases:
                return groups, bases
            else:
                return groups

        # Sorted insertion heuristic https://quantum-journal.org/papers/q-2021-01-20-385/pdf/
        sorted_terms = sorted(self.terms_dict.items(), key=lambda item: abs(item[1]), reverse=True)

        for term,coeff in sorted_terms:

            commute_bool = False
            if len(groups)>0:
                n = len(groups)
                for i in range(n):
                    commute_bool = bases[i].commute_qw(term)
                    if commute_bool:
                        bases[i].update(term.factor_dict)
                        groups[i].terms_dict[term]=coeff
                        break
            if len(groups)==0 or not commute_bool:
                groups.append(QubitOperator({term:coeff}))
                bases.append(term.copy())

        if show_bases:
            return groups, bases
        else:
            return groups
    
    #
    # Measurement settings and measurement
    #
    
    def change_of_basis(self, qarg, method="commuting_qw"):
        """
        Performs several operations on a quantum argument such that the hermitian
        part of self is diagonal when conjugated with these operations.

        Parameters
        ----------
        qarg : QuantumVariable or list[Qubit]
            The quantum argument to apply the change of basis on.
        method : str, optional
            The method for calculating the change of basis. 
            Available are ``commuting`` (all QubitTerms must mutually commute) and ``commuting_qw`` (all QubitTerms must mutually commute qubit-wise).
            The default is ``commuting_qw``.

        Returns
        -------
        res : QubitOperator
            A qubit operator that contains only diagonal entries (I, Z, P0, P1).

        """
        
        # Assuming all terms of self commute qubit-wise,
        # the basis change for Pauli factor is trivial:
        # Z stays the same, for X we apply an h gate and for Y and s_dg.
        
        # For ladder operators, the situation is more intricate.
        
        # Take for instance the ladder operators A(0)*A(1)*A(2) + h.c.
        
        # In Bra-Ket form, this is |000><111| + |111><000|
        
        # The considerations from Selingers Paper https://arxiv.org/abs/2310.12256
        
        # In this work, the above term is simulated by the following circuit
        
        #                ┌───┐                                                        ┌───┐
        #   qv_0.0: ─────┤ X ├────────────■─────────────────────────■─────────────────┤ X ├─────
        #                └─┬─┘┌───┐       │                         │            ┌───┐└─┬─┘
        #   qv_0.1: ───────┼──┤ X ├───────■─────────────────────────■────────────┤ X ├──┼───────
        #           ┌───┐  │  └─┬─┘┌───┐  │  ┌───┐┌──────────────┐  │  ┌───┐┌───┐└─┬─┘  │  ┌───┐
        #   qv_0.2: ┤ X ├──■────■──┤ X ├──┼──┤ H ├┤ Rz(-1.0*phi) ├──┼──┤ H ├┤ X ├──■────■──┤ X ├
        #           └───┘          └───┘┌─┴─┐└───┘└──────┬───────┘┌─┴─┐└───┘└───┘          └───┘
        # hs_anc.0: ────────────────────┤ X ├────────────■────────┤ X ├─────────────────────────
        #                               └───┘                     └───┘
        
        
        # From this we conclude that H can be expressed as a conjugation of the following form.
        
        # H = U^dg (|110><110| - |111><111|)/2 U
        
        # Where U is the following circuit:
        
        #             ┌───┐               
        # qb_90: ─────┤ X ├───────────────
        #             └─┬─┘┌───┐          
        # qb_91: ───────┼──┤ X ├──────────
        #        ┌───┐  │  └─┬─┘┌───┐┌───┐
        # qb_92: ┤ X ├──■────■──┤ X ├┤ H ├
        #        └───┘          └───┘└───┘
                        
        
        # This is because
        
        # exp(i*t*H) = U^dg MCRZ(i*t) U
        #            = U^dg exp(i*t*(|110><110| - |111><111|)/2) U
        
        # The bra-ket term is already diagonal but how to express it via operators?
        
        # The answer is P1(0)*P1(1)*Z(2)
        
        # From this we conclude the underlying rule here. For ladder terms we can
        # pick an arbitrary qubit that we call "anchor qubit" which is conjugated
        # with an H gate.
        
        # After performing the conjugation with the CX gates to complete the inverse
        # GHZ preparation, the ladder operator transforms into a chain of projectors
        # whereas the anchor qubit becomes a Z gate.
        
        n = self.find_minimal_qubit_amount()
        if len(qarg) < n:
            raise Exception("Tried to change the basis of an Operator on a quantum argument with insufficient qubits.")
        
     
        # This dictionary will contain the new terms/coefficient comination for the
        # diagonal operator
        new_terms_dict = {}

        new_factor_dicts = []
        prefactors = []
        
        ladder_conjugation_performed = False
        ladder_indices = []

        if method=="commuting_qw":
            # We track which qubit is in which basis to raise an error if a
            # violation with the requirement of qubit wise commutativity is detected.
            basis_dict = {}            
    
            # We iterate through the terms and apply the appropriate basis transformation
            for term, coeff in self.terms_dict.items():

                factor_dict = term.factor_dict
                # This dictionary will contain the factors of the new term
                new_factor_dict = {}
                
                new_factor_dicts.append(new_factor_dict)
            
                prefactor = 1
                prefactors.append(prefactor)
            
                for j in range(n):
                
                    # If there is no entry in the factor dict, this corresponds to
                    # identity => no basis change required.
                    if j not in factor_dict:
                        continue
                
                    # If j is already in the basis dict, we assert that the bases agree
                    # (otherwise there is a violation of qubit-wise commutativity)
                    if j in basis_dict:
                        if basis_dict[j] != factor_dict[j]:
                            assert basis_dict[j] in ["Z", "P0", "P1"]
                        new_factor_dict[j] = "Z"
                        continue
                    
                    # We treat ladder operators in the next section
                    if factor_dict[j] not in ["X", "Y", "Z"]:
                        continue
                
                    # Update the basis dict
                    basis_dict[j] = factor_dict[j]
                
                    # Append the appropriate basis-change gate
                    if factor_dict[j]=="X":
                        h(qarg[j])
                    
                    if factor_dict[j]=="Y":
                        sx_dg(qarg[j])
            
                    new_factor_dict[j] = "Z"
                    
                    

        if method=="commuting":

            # Calculate S: Matrix where the colums correspond to the binary representation (Z/X) of the Pauli terms
            x_vectors = []
            z_vectors = []
            for term, coeff in self.terms_dict.items():
                x_vector, z_vector = term.binary_representation(n)
                x_vectors.append(x_vector)
                z_vectors.append(z_vector)
            x_matrix = np.stack(x_vectors, axis=1)
            z_matrix = np.stack(z_vectors, axis=1)

            # Find qubits (rows) on which Pauli X,Y,Z operatos act
            qb_indices = []
            for k in range(n):
                if not (np.all(x_matrix[k] == 0) and np.all(z_matrix[k] == 0)):
                    qb_indices.append(k)
            m = len(qb_indices)
            
            if m==0:
                new_factor_dicts = [{} for _ in range(self.len())]
                prefactors = [1]*self.len()
            else:
                S = np.vstack((z_matrix[qb_indices], x_matrix[qb_indices]))
            
                # Construct and apply change of basis
                A, R_inv, h_list, s_list, perm = construct_change_of_basis(S)

                def inv_graph_state(qarg):
                    for i in range(m):
                        for j in range(i):
                            if A[i,j]==1:
                                cz(qarg[qb_indices[perm[i]]],qarg[qb_indices[perm[j]]])
                    for i in qb_indices:
                        h(qarg[i])

                def change_of_basis(qarg):
                    for i in h_list:
                        h(qarg[qb_indices[i]])
                    for i in s_list:
                        s(qarg[qb_indices[perm[i]]])
                    inv_graph_state(qarg)

                change_of_basis(qarg)

                # Construct new QubitOperator
                #
                # Factor (-1) appears if S gate is applied to X, or Hadamard gate H is applied to Y:
                # S^dagger X S = -Y
                # S^dagger Y S = X
                # S^dagger Z S = Z
                # H X H = Z
                # H Y H = -Y
                # H Z H = X
                # For the original Pauli terms this translates to: Factor (-1) appears if S gate is applied to Y, or Hadamard gate H is applied to Y
                # No factor (-1) occurs if H S^{-1} P S H is applied (i.e., H and S) for any P in {X,Y,Z}

                s_vector = np.zeros(m, dtype=int)
                s_vector[s_list] = 1
                h_vector = np.zeros(m, dtype=int)
                h_vector[h_list] = 1
                sh_vector = s_vector[perm] + h_vector % 2 
                sign_vector = sh_vector @ (x_matrix[qb_indices]*z_matrix[qb_indices]) % 2

                # Lower triangular part of A
                A_low = np.tril(A)

                for index,z_vector in enumerate(R_inv.T):

                    # Determine the sign of the product of the selected graph state stabilizers:
                    # 
                    # Consider product of stabilizers S_{i_1}*S_{i_2}*...*S_{i_m} with (w.l.o.g.) i_1<i_2<...<i_m
                    # For each i: Swap X_i with all Z_i's from stabilizers if index > i such that all Z_i's are on the left of X_i
                    # Calculate the paritiy n1 of the sum of the numbers of 1's with position j>i for each row of the square submatrix A defined by z_vector
                    # Yields a factor (-1)^n1

                    n1 = sum((z_vector @ A_low)*z_vector) % 2

                    # For each i: Count the number of Z_i's: if even, no factor, if odd: factor i (ZX=iY)
                    # Count the number n2 of rows of the square submatrix of A defined by z_vector, such that the number of 1's in each row is odd
                    # This number is always even since A is a symmetric matrix with 0's on the diagonal
                    # Yields a factor i^n2=(-1)^(n2/2)

                    n2 = sum((z_vector @ A)*z_vector % 2)

                    new_factor_dict = {qb_indices[perm[i]]:"Z" for i in range(m) if z_vector[i]==1}
                    new_factor_dicts.append(new_factor_dict)
                    
                    prefactor = (-1)**sign_vector[index]*(-1)**(n1+n2/2)
                    prefactors.append(prefactor)

        # Ladder operators 
        for term, coeff in self.terms_dict.items():    

            prefactor = prefactors.pop(0)    
            new_factor_dict = new_factor_dicts.pop(0)
            
            # Next we treat the ladder operators
            ladder_operators = [base for base in term.factor_dict.items() if base[1] in ["A", "C"]]
            
            if len(ladder_operators):
                
                # The anchor factor is the "last" ladder operator. 
                # This is the qubit where the H gate will be executed.
                anchor_factor = ladder_operators[-1]
                new_factor_dict[ladder_operators[-1][0]] = "Z"
                
                # Perform the cnot gates
                for j in range(len(ladder_operators)-1):
                    if not ladder_conjugation_performed:
                        ladder_indices.append(ladder_operators[j][0])
                        cx(qarg[anchor_factor[0]], qarg[ladder_operators[j][0]])
                    
                    if anchor_factor[1] == "C":
                        if ladder_operators[j][1] == "A":
                            new_factor_dict[ladder_operators[j][0]] = "P1"
                        else:
                            new_factor_dict[ladder_operators[j][0]] = "P0"
                    else:
                        if ladder_operators[j][1] == "A":
                            new_factor_dict[ladder_operators[j][0]] = "P0"
                        else:
                            new_factor_dict[ladder_operators[j][0]] = "P1"
                
                if not ladder_conjugation_performed:
                    # Execute the H-gate
                    ladder_indices.append(anchor_factor[0])
                    h(qarg[anchor_factor[0]])
                else:
                    if set(ladder_indices) != set(ladder_factor[0] for ladder_factor in ladder_operators):
                        raise Exception("Tried to perform change of basis on operator containing non-matching ladder indices")
                
                ladder_conjugation_performed = True
                
                prefactor *= 0.5
                
            for k, v in term.factor_dict.items():
                if v in ["P0", "P1"]:
                    new_factor_dict[k] = v

            new_term = QubitTerm(new_factor_dict)
            new_terms_dict[new_term] = prefactor*self.terms_dict[term]
        
        return QubitOperator(new_terms_dict) 
        
    
    def get_conjugation_circuit(self):
        # This method returns a QuantumCircuit that should be applied
        # before a measurement of self is peformed.
        # The method assumes that all terms within this Operator commute qubit-
        # wise. For instance, if an X operator is supposed to be measured,
        # the conjugation circuit will contain an H gate at that point,
        # because the X operator can be measured by measuring the Z Operator
        # in the H-transformed basis.
        
        # For the ladder operators, the conjugation circuit not this straight-
        # forward. To understand how we measure the ladder operators, consider
        # the operator
        
        # H = (A(0)*A(1)*A(2) + h.c.)
        #   = (|000><111| + |111><000|)
        
        # The considerations from Selingers Paper https://arxiv.org/abs/2310.12256
        # motivate that H can be expressed as a conjugation of the following form.
        
        # H = U^dg (|110><110| - |111><111|)/2 U
        
        # This is because
        
        # exp(i*t*H) = U^dg MCRZ(i*t) U
        #            = U^dg exp(i*t*(|110><110| - |111><111|)/2) U
        
        # We use this insight because the Operator 
        # |111><111| - |110><110| = |11><11| (x) (|0><0| - |1><1|) 
        # = |11><11| (x) Z
        # can be measured via postprocessing.
        
        # The postprocessing to do is essentially measuring the last qubit as
        # a regular Z operator and only add the result to the expectation value
        # if the first two qubits are measured to be in the |1> state.
        # If they are in any other state nothing should be added.
        
        # From this we can also conclude how the conjugation circuit needs to
        # look like: Essentially like the conjugation circuit from the paper.
        
        # For our example above (when simulated) gives:
            
        #                ┌───┐                                                        ┌───┐
        #   qv_0.0: ─────┤ X ├────────────■─────────────────────────■─────────────────┤ X ├─────
        #                └─┬─┘┌───┐       │                         │            ┌───┐└─┬─┘
        #   qv_0.1: ───────┼──┤ X ├───────■─────────────────────────■────────────┤ X ├──┼───────
        #           ┌───┐  │  └─┬─┘┌───┐  │  ┌───┐┌──────────────┐  │  ┌───┐┌───┐└─┬─┘  │  ┌───┐
        #   qv_0.2: ┤ X ├──■────■──┤ X ├──┼──┤ H ├┤ Rz(-1.0*phi) ├──┼──┤ H ├┤ X ├──■────■──┤ X ├
        #           └───┘          └───┘┌─┴─┐└───┘└──────┬───────┘┌─┴─┐└───┘└───┘          └───┘
        # hs_anc.0: ────────────────────┤ X ├────────────■────────┤ X ├─────────────────────────
        #                               └───┘                     └───┘
        
        # Where the construction of the MCRZ gate is is encoded into the Toffolis
        # and the controlled RZ-Gate.
        
        # The conjugation circuit therefore needs to look like this:
                    
        #             ┌───┐               
        # qb_90: ─────┤ X ├───────────────
        #             └─┬─┘┌───┐          
        # qb_91: ───────┼──┤ X ├──────────
        #        ┌───┐  │  └─┬─┘┌───┐┌───┐
        # qb_92: ┤ X ├──■────■──┤ X ├┤ H ├
        #        └───┘          └───┘└───┘
                
        # To learn more about how the post-processing is implemented check the
        # comments of QubitTerm.serialize
        
        # ===============
        
        # Create a QuantumCircuit that contains the conjugation
        from qrisp import QuantumCircuit
        n = self.find_minimal_qubit_amount()
        qc = QuantumCircuit(n)
        
        # We track which qubit is in which basis to raise an error if a
        # violation with the requirement of qubit wise commutativity is detected.
        basis_dict = {}
        
        # We iterate through the terms and apply the appropriate basis transformation
        for term, coeff in self.terms_dict.items():
            
            factor_dict = term.factor_dict
            for j in range(n):
                
                # If there is no entry in the factor dict, this corresponds to
                # identity => no basis change required.
                if j not in factor_dict:
                    continue
                
                # If j is already in the basis dict, we assert that the bases agree
                # (otherwise there is a violation of qubit-wise commutativity)
                if j in basis_dict:
                    assert basis_dict[j] == factor_dict[j]
                    continue
                
                # We treat ladder operators in the next section
                if factor_dict[j] not in ["X", "Y", "Z"]:
                    continue
                
                # Update the basis dict
                basis_dict[j] = factor_dict[j]
                
                # Append the appropriate basis-change gate
                if factor_dict[j]=="X":
                    qc.h(j)
                if factor_dict[j]=="Y":
                    qc.sx(j)
            
            # Next we treat the ladder operators
            ladder_operators = [base for base in term.factor_dict.items() if base[1] in ["A", "C"]]
            
            if len(ladder_operators):
                
                # The anchor factor is the "last" ladder operator. 
                # This is the qubit where the H gate will be executed.
                anchor_factor = ladder_operators[-1]
            
                # Flip the anchor qubit if the ladder operator is an annihilator
                if anchor_factor[1] == "C":
                    qc.x(anchor_factor[0])
                
                # Perform the cnot gates
                for j in range(len(ladder_operators)-1):
                    qc.cx(anchor_factor[0], ladder_operators[j][0])

                # Flip the anchor qubit back
                if anchor_factor[1] == "C":
                    qc.x(anchor_factor[0])
            
                # Execute the H-gate
                qc.h(anchor_factor[0])
        
        return qc, QubitOperator(self.terms_dict)
    
    def get_operator_variance(self, n = 1):
        """
        Calculates the optimal distribution and number of shots following https://quantum-journal.org/papers/q-2021-01-20-385/pdf/.
        
        Normally to compute the variance of an operator, the distribution has to be known.
        Since the distribution is not known without querying the quantum device,
        the authors estimate the variance as the expectation value of a distribution
        of quantum states. This distribution is uniform across the unit sphere.
        
        For an arbitrary Pauli-Operator P != I they conclude
        
        E(Var(P)) = alpha_n = 1 - 1/(2^n + 1)
        
        Where 2^n is the dimension of the comprising space
        
        Since the QubitOperator class also contains A, C and P operators, we have to
        do more work.
        
        To understand how the variance can be estimated, recall that every
        QubitOperator O can be transformed to a sum of Pauli strings
        
        Var(O) = Var(sum_i(c_i*P_i))
               = sum_i(Var(c_i*P_i)) + 2*sum_[0<=i<j<=n](Cov(c_i*P_i,c_j*P_j))
        
        The last line can be found in https://arxiv.org/pdf/1907.13623 section 10.1.
        Theorem 2 of that very same source states that for the above distribution
        of states, we have E(Cov(P_i, P_j)) = 0 if P_i != P_j
        
        From that we conclude
        
        E(Var(O)) = sum_i(E(Var(c_i*P_i)))
                  = sum_i(abs(c_i)**2*E(Var(P_i)))
                  = alpha_n * sum_i(abs(c_i)**2)
                  
        It therefore suffices to compute the variance of the Pauli form of the
        QubitOperator.
        
        
        """
        var = 0
        pauli_form = self.hermitize().to_pauli()
        for term, coeff in pauli_form.terms_dict.items():
            if len(term.factor_dict) != 0:
                var += abs(coeff)**2
        alpha_n = 1 - 1/(2**n + 1)
        
        return var*alpha_n
        
    def get_measurement(
        self,
        qarg,
        precision=0.01,
        backend=None,
        shots=1000000,
        compile=True,
        compilation_kwargs={},
        subs_dic={},
        precompiled_qc=None,
        diagonalisation_method="commuting_qw",
        measurement_data=None # measurement settings
    ):
        r"""
        This method returns the expected value of a Hamiltonian for the state 
        of a quantum argument. Note that this method measures the **hermitized**
        version of the operator:
            
        .. math::
            
            H = (O + O^\dagger)/2


        Parameters
        ----------
        qarg : QuantumVariable or list[Qubit]
            The quantum argument to evaluate the Hamiltonian on.
        precision: float, optional
            The precision with which the expectation of the Hamiltonian is to be evaluated.
            The default is 0.01. The number of shots scales quadratically with the inverse precision.
        backend : BackendClient, optional
            The backend on which to evaluate the quantum circuit. The default can be
            specified in the file default_backend.py.
        shots : integer, optional
            The maximum amount of shots to evaluate the expectation of the Hamiltonian. 
            The default is 1000000.
        compile : bool, optional
            Boolean indicating if the .compile method of the underlying QuantumSession
            should be called before. The default is True.
        compilation_kwargs  : dict, optional
            Keyword arguments for the compile method. For more details check
            :meth:`QuantumSession.compile <qrisp.QuantumSession.compile>`. The default
            is ``{}``.
        subs_dic : dict, optional
            A dictionary of Sympy symbols and floats to specify parameters in the case
            of a circuit with unspecified, :ref:`abstract parameters<QuantumCircuit>`.
            The default is {}.
        precompiled_qc : QuantumCircuit, optional
            A precompiled quantum circuit.
        measurement_data : QubitOperatorMeasurement
            Cached data to accelerate the measurement procedure. Automatically generated by default.

        Raises
        ------
        Exception
            If the containing QuantumSession is in a quantum environment, it is not
            possible to execute measurements.

        Returns
        -------
        float
            The expected value of the Hamiltonian.

        Examples
        --------

        We define a Hamiltonian, and measure its expected value for the state of a :ref:`QuantumVariable`.

        ::

            from qrisp import QuantumVariable, h
            from qrisp.operators.qubit import X,Y,Z
            qv = QuantumVariable(2)
            h(qv)
            H = Z(0)*Z(1)
            res = H.get_measurement(qv)
            print(res)
            #Yields 0.0011251406425802912

        """
        return get_measurement(self, 
                                qarg, 
                                precision=precision, 
                                backend=backend, 
                                shots=shots, 
                                compile=compile, 
                                compilation_kwargs=compilation_kwargs, 
                                subs_dic=subs_dic,
                                precompiled_qc=precompiled_qc, 
                                diagonalisation_method=diagonalisation_method,
                                measurement_data=measurement_data)

    #
    # Trotterization
    #

    def trotterization(self, method='commuting_qw'):
        r"""
        .. _ham_sim:
        
        Returns a function for performing Hamiltonian simulation, i.e., approximately implementing the unitary operator $e^{itH}$ via Trotterization.
        Note that this method will always simulate the **hermitized** operator, i.e.
        
        .. math::
            
            H = (O + O^\dagger)/2


        Parameters
        ----------
        method : str, optional
            The method for grouping the QubitTerms. 
            Available are ``commuting`` (groups such that all QubitTerms mutually commute) and ``commuting_qw`` (groups such that all QubitTerms mutually commute qubit-wise).
            The default is ``commuting_qw``.

        Returns
        -------
        U : function 
            A Python function that implements the first order Suzuki-Trotter formula.
            Given a Hamiltonian $H=H_1+\dotsb +H_m$ the unitary evolution $e^{itH}$ is 
            approximated by 
            
            .. math::

                e^{itH}\approx U_1(t,N)=\left(e^{iH_1t/N}\dotsb e^{iH_mt/N}\right)^N

            This function receives the following arguments:

            * qarg : QuantumVariable 
                The quantum argument.
            * t : float, optional
                The evolution time $t$. The default is 1.
            * steps : int, optional
                The number of Trotter steps $N$. The default is 1.
            * iter : int, optional 
                The number of iterations the unitary $U_1(t,N)$ is applied. The default is 1.
                
        Examples
        --------
        
        We simulate a simple QubitOperator.
        
        >>> from sympy import Symbol
        >>> from qrisp.operators import A,C,Z,Y
        >>> from qrisp import QuantumVariable
        >>> O = A(0)*C(1)*Z(2) + Y(3)
        >>> U = O.trotterization()
        >>> qv = QuantumVariable(4)
        >>> t = Symbol("t")
        >>> U(qv, t = t)
        >>> print(qv.qs)
        QuantumCircuit:
        ---------------
              ┌───┐                       ┌───┐┌───────────┐ ┌───┐          ┌───┐
        qv.0: ┤ X ├───────────────────────┤ X ├┤ Rz(0.5*t) ├─┤ X ├──────────┤ X ├
              └─┬─┘     ┌───┐      ┌───┐  └─┬─┘├───────────┴┐└─┬─┘┌───┐┌───┐└─┬─┘
        qv.1: ──■───────┤ H ├──────┤ X ├────■──┤ Rz(-0.5*t) ├──■──┤ X ├┤ H ├──■──
                        └───┘      └─┬─┘       └────────────┘     └─┬─┘└───┘     
        qv.2: ───────────────────────■──────────────────────────────■────────────
              ┌────┐┌────────────┐┌──────┐                                       
        qv.3: ┤ √X ├┤ Rz(-2.0*t) ├┤ √Xdg ├───────────────────────────────────────
              └────┘└────────────┘└──────┘                                       
        Live QuantumVariables:
        ----------------------
        QuantumVariable qv
        
        Execute a simulation:
            
        >>> print(qv.get_measurement(subs_dic = {t : 0.5}))
        {'0000': 0.77015, '0001': 0.22985}
        
        """
        O = self.hermitize().eliminate_ladder_conjugates()
        commuting_groups = O.group_up(lambda a, b: a.commute(b))
        
        if method=='commuting_qw':
            def trotter_step(qarg, t, steps):
                for com_group in commuting_groups:
                    qw_groups, bases = com_group.commuting_qw_groups(show_bases=True)
                    for index,basis in enumerate(bases):
                        qw_group = qw_groups[index]
                        with conjugate(qw_group.change_of_basis)(qarg) as diagonal_operator:
                            intersect_groups = diagonal_operator.group_up(lambda a, b: not a.intersect(b))
                            for intersect_group in intersect_groups:
                                for term,coeff in intersect_group.terms_dict.items():
                                    term.simulate(coeff*t/steps, qarg, do_change_of_basis = False)
        
        if method=='commuting':
            def trotter_step(qarg, t, steps):
                for com_group in commuting_groups:
                    with conjugate(com_group.change_of_basis)(qarg,method="commuting") as diagonal_operator:
                        intersect_groups = diagonal_operator.group_up(lambda a, b: not a.intersect(b))
                        for intersect_group in intersect_groups:
                            for term,coeff in intersect_group.terms_dict.items():
                                term.simulate(coeff*t/steps, qarg)

        def U(qarg, t=1, steps=1, iter=1):
            merge([qarg])
            with IterationEnvironment(qarg.qs, iter*steps):
                trotter_step(qarg, t, steps)

        return U

